from __future__ import annotations
import jax.numpy as np
from jax import vmap
from jax.tree_util import tree_map
from jax.lax import stop_gradient
from zodiax import Base
from abc import ABC, abstractmethod
from inspect import signature
from dLux.utils.helpers import two_image_plot
from dLux.utils.units import convert_angular, convert_cartesian
import dLux


__all__ = ["CreateWavefront", "TiltWavefront", "NormaliseWavefront", 
           "ApplyBasisOPD", "AddPhase", "AddOPD", "TransmissiveOptic", 
           "ApplyBasisCLIMB", "Rotate"]


Array = np.ndarray


class OpticalLayer(Base, ABC):
    """
    A base Optical layer class to help with type checking throuhgout the rest
    of the software. Instantiates the apply method which inspects the function
    signature of the __call__ method in order to only pass and return the
    parameters dictionary if it is needed and modified.

    Attributes
    ----------
    name : str
        The name of the layer, which is used to index the layers dictionary.
    """
    name : str


    def __init__(self : OpticalLayer,
                 name : str = 'OpticalLayer') -> OpticalLayer:
        """
        Constructor for the OpticalLayer class.

        Parameters
        ----------
        name : str = 'OpticalLayer'
            The name of the layer, which is used to index the layers dictionary.
        """
        self.name = str(name)


    @abstractmethod
    def __call__(self : OpticalLayer, wavefront : Wavefront) -> Wavefront: # pragma: no cover
        """
        Appies the layer to the `Wavefront`.

        Parameters
        ----------
        wavefront : Wavefront
            The wavefront to operate on.

        Returns
        -------
        wavefront : Wavefront
            The wavefront with the optical layer applied.
        """
        return


    def apply(self : OpticalLayer, parameters : dict) -> dict:
        """
        Unpacks the wavefront object from the parameters dictionary and applies
        the layers __call__ method upon it.

        Parameters
        ----------
        parameters : dict
            A dictionary that must contain a "Wavefront" key with a
            corresponding dLux.wavefronts.Wavefront object.

        Returns
        -------
        parameters : dict
            A dictionary with the updated "Wavefront" key with the propagated
            wavefront object.
        """
        # Inspect apply function to see if it takes/returns the parameters dict
        input_parameters = signature(self.__call__).parameters

        # Method does not take in the parameters, update in place
        if 'parameters' not in input_parameters:
            parameters["Wavefront"] = self.__call__(parameters["Wavefront"])

        # Method takes and return updated parameters
        elif input_parameters['returns_parameters'].default == True:
            wavefront, parameters = self.__call__(parameters["Wavefront"],
                                                  parameters)
            parameters["Wavefront"] = wavefront

        # Method takes but does not return parameters
        else:
            parameters["Wavefront"] = self.__call__(parameters["Wavefront"],
                                                    parameters)

        # Return updated parameters dictionary
        return parameters
    

    def summary(self            : OpticalLayer, 
                angular_units   : str = 'radians', 
                cartesian_units : str = 'meters', 
                sigfigs         : int = 4) -> str:
        """
        Returns a summary of the class.

        Parameters
        ----------
        angular_units : str = 'radians'
            The angular units to use in the summary. Options are 'radians', 
            'degrees', 'arcseconds' and 'arcminutes'.
        cartesian_units : str = 'meters'
            The cartesian units to use in the summary. Options are 'meters',
            'millimeters' and 'microns'.
        sigfigs : int = 4
            The number of significant figures to use in the summary.

        Returns
        -------
        summary : str
            A summary of the class.
        """
        return f"{self.name} layer has no summary method yet."


    def display(self            : OpticalLayer, 
                wavefront       : Wavefront,
                figsize         : tuple = (10, 4),
                dpi             : int = 120,
                angular_units   : str = 'radians', 
                cartesian_units : str = 'meters', 
                sigfigs         : int = 4) -> None:
        """
        Displays a plot of the wavefront amplitude and opd or phase.

        Parameters
        ----------
        wavefront : Wavefront
            The dummy wavefront to propagate though the optics.
        figsize : tuple = (10, 4)
            The size of the figure to display.
        cmap : str = 'inferno'
            The colour map to use.
        dpi : int = 120
            The resolution of the figure.
        angular_units : str = 'radians'
            The angular units to use in the summary. Options are 'radians', 
            'degrees', 'arcseconds' and 'arcminutes'.
        cartesian_units : str = 'meters'
            The cartesian units to use in the summary. Options are 'meters',
            'millimeters' and 'microns'.
        sigfigs : int = 4
            The number of significant figures to use in the summary.
        """
        amplitude = wavefront.amplitude
        phase = convert_angular(wavefront.phase, "radians", angular_units)
        two_image_plot(amplitude[0], phase[0], figsize=figsize, 
            titles=("Amplitude", "Phase"), cbar_labels=("Intensity", 
            f"Phase ({angular_units})"), cmaps=('inferno', 'twilight'), 
            bounds=(None, 2*np.pi), dpi=dpi)


class CreateWavefront(OpticalLayer):
    """
    Initialises the relevant Wavefront class with the specified attributes.
    Also applies the tilt specified by the source object, defined in the
    parameters dictionary. All wavefronts are cosntructed in the Pupil plane.

    Attributes
    ----------
    npixels : int
        The number of pixels used to represent the wavefront.
    diameter: Array, meters
        The diameter of the wavefront in the Pupil plane.
    wavefront_type: str
        Determines the type of wavefront class to create. Currently supports
        'Cartesian', 'Angular', 'FarFieldFresnel'.
    name : str
        The name of the layer, which is used to index the layers dictionary.
    """
    npixels        : int
    diameter       : Array
    wavefront_type : str


    def __init__(self           : OpticalLayer,
                 npixels        : int,
                 diameter       : Array,
                 wavefront_type : str = 'Angular',
                 name           : str = 'CreateWavefront') -> OpticalLayer:
        """
        Constructor for the CreateWavefront class.

        Parameters
        ----------
        npixels : int
            The number of pixels used to represent the wavefront.
        diameter: Array, meters
            The diameter of the wavefront in the Pupil plane.
        wavefront_type: str = 'Angular'
            Determines the type of wavefront class to create. Currently supports
            'Cartesian', 'Angular', 'FarFieldFresnel'.
        name : str = 'CreateWavefront'
            The name of the layer, which is used to index the layers dictionary.
            Default is 'CreateWavefront'.
        """
        super().__init__(name)
        self.npixels        = int(npixels)
        self.diameter       = np.asarray(diameter, dtype=float)
        self.wavefront_type = str(wavefront_type)

        # Input checks
        assert self.diameter.ndim == 0, ("diameter must be "
        "a scalar array.")
        assert wavefront_type in ('Cartesian', 'Angular', 'FarFieldFresnel'), \
        ("wavefront_type must be either 'Cartesian', 'Angular' or "
         "'FarFieldFresnel'")


    def __call__(self               : OpticalLayer,
                 wavefront          : Wavefront,
                 parameters         : dict,
                 returns_parameters : bool = True) -> Wavefront:
        """
        Constructs a wavefront obect based on the parameters of the class and
        the parameters within the parameters dictionary.

        Parameters
        ----------
        wavefront : None
            Any empty None type input to the class in order to maintain the
            input conventions determied by the apply method of OpticalLayers.
        parameters : dict
            A dictionary of parameters needed to construct the wavefront.
        returns_parameters : bool = True
            Determines if the class returns the parameters dictionary.

        Returns
        -------
        wavefront, parameters : (Wavefront, dict)
            Returns the constructed wavefront and the updated parameters
            dictionary. If returns_parameters is False, only the wavefront is
            returned.
        """
        # Get the wavelength
        wavelength = parameters["wavelength"]

        # Determine the pixel scale
        pixel_scale = self.diameter/self.npixels

        # Construct normalised Amplitude
        amplitude = np.ones((1, self.npixels, self.npixels))
        amplitude /= np.linalg.norm(amplitude)

        # Construct empty phases
        phase = np.zeros(((1, self.npixels, self.npixels)))

        # Get correct Wavefront type
        wavefront_constructor = getattr(dLux.wavefronts,
                                        self.wavefront_type + "Wavefront")

        # Construct Wavefront
        wavefront = wavefront_constructor(wavelength, pixel_scale, amplitude,
                                      phase, dLux.wavefronts.PlaneType.Pupil)

        # Tilt wavefront from source offset
        wavefront = wavefront.tilt_wavefront(parameters["offset"])

        # Kill PlaneType Gradients
        is_leaf = lambda x: isinstance(x, dLux.wavefronts.PlaneType)
        kill_gradient = lambda x: stop_gradient(x.value) if is_leaf(x) else x
        wavefront = tree_map(kill_gradient, wavefront, is_leaf=is_leaf)

        # Update the parameters dictionary with the constructed wavefront
        parameters["Wavefront"] = wavefront

        # Return either the wavefront or wavefront and parameters dictionary
        if returns_parameters:
            return wavefront, parameters
        else:
            return wavefront
    

    def summary(self            : OpticalLayer, 
                angular_units   : str = 'radians', 
                cartesian_units : str = 'meters', 
                sigfigs         : int = 4) -> str:
        """
        Returns a summary of the class.

        Parameters
        ----------
        angular_units : str = 'radians'
            The angular units to use in the summary. Options are 'radians', 
            'degrees', 'arcseconds' and 'arcminutes'.
        cartesian_units : str = 'meters'
            The cartesian units to use in the summary. Options are 'meters',
            'millimeters' and 'microns'.
        sigfigs : int = 4
            The number of significant figures to use in the summary.

        Returns
        -------
        summary : str
            A summary of the class.
        """
        return (f"{self.name}: Creates a {self.wavefront_type} wavefront of " 
                f"size {self.npixels} pixels and diameter {self.diameter} m.")


class TiltWavefront(OpticalLayer):
    """
    Tilts the wavefront by the input tilt_angles.

    Attributes
    ----------
    tilt_angles : Array, radians
        The (x, y) angles by which to tilt the wavefront.
    name : str
        The name of the layer, which is used to index the layers dictionary.
    """
    tilt_angles : Array


    def __init__(self        : OpticalLayer,
                 tilt_angles : Array,
                 name        : str = 'TiltWavefront') -> OpticalLayer:
        """
        Constructor for the TiltWavefront class.

        Parameters
        ----------
        tilt_angles : Array, radians
            The (x, y) angles by which to tilt the wavefront.
        name : str = TiltWavefront
            The name of the layer, which is used to index the layers dictionary.
            Default is 'TiltWavefront'.
        """
        super().__init__(name)
        self.tilt_angles = np.asarray(tilt_angles, dtype=float)

        # Input checks
        assert self.tilt_angles.shape == (2,), \
        ("tilt_angles must be an array of shape (2,), ie (x, y).")


    def __call__(self : OpticalLayer, wavefront : Wavefront) -> Wavefront:
        """
        Applies the tilt_angle to the phase of the wavefront.

        Parameters
        ----------
        wavefront : Wavefront
            The wavefront to operate on.

        Returns
        -------
        wavefront : Wavefront
            The wavefront with the optical layer applied.
        """
        return wavefront.tilt_wavefront(self.tilt_angles)


    def summary(self            : OpticalLayer, 
                angular_units   : str = 'radians', 
                cartesian_units : str = 'meters', 
                sigfigs         : int = 4) -> str:
        """
        Returns a summary of the class.

        Parameters
        ----------
        angular_units : str = 'radians'
            The angular units to use in the summary. Options are 'radians', 
            'degrees', 'arcseconds' and 'arcminutes'.
        cartesian_units : str = 'meters'
            The cartesian units to use in the summary. Options are 'meters',
            'millimeters' and 'microns'.
        sigfigs : int = 4
            The number of significant figures to use in the summary.

        Returns
        -------
        summary : str
            A summary of the class.
        """
        angles = convert_angular(self.tilt_angles, 'radians', angular_units)
        return (f"{self.name}: Tilts the wavefront by {angles:.{sigfigs}} "
                f"{angular_units} in the (x, y) dimension.")


class NormaliseWavefront(OpticalLayer):
    """
    Normalises the input wavefront using the in-built wavefront normalisation
    method.

    Attributes
    ----------
    name : str
        The name of the layer, which is used to index the layers dictionary.
    """
    

    def __init__(self : OpticalLayer,
                 name : str = 'NormaliseWavefront') -> OpticalLayer:
        """
        Constructor for the NormaliseWavefront class.

        Parameters
        ----------
        name : string = 'NormaliseWavefront'
            The name of the layer, which is used to index the layers
            dictionary.
        """
        super().__init__(name)


    def __call__(self : OpticalLayer, wavefront : Wavefront) -> Wavefront:
        """
        Normalises the wavefront.

        Parameters
        ----------
        wavefront : Wavefront
            The wavefront to operate on.

        Returns
        -------
        wavefront : Wavefront
            The wavefront with the wavefront normalisation method applied.
        """
        return wavefront.normalise()
    

    def summary(self            : OpticalLayer, 
                angular_units   : str = 'radians', 
                cartesian_units : str = 'meters', 
                sigfigs         : int = 4) -> str:
        """
        Returns a summary of the class.

        Parameters
        ----------
        angular_units : str = 'radians'
            The angular units to use in the summary. Options are 'radians', 
            'degrees', 'arcseconds' and 'arcminutes'.
        cartesian_units : str = 'meters'
            The cartesian units to use in the summary. Options are 'meters',
            'millimeters' and 'microns'.
        sigfigs : int = 4
            The number of significant figures to use in the summary.

        Returns
        -------
        summary : str
            A summary of the class.
        """
        return f"{self.name}: Normalises the wavefront to unity power."


class ApplyBasisOPD(OpticalLayer):
    """
    Adds an array of phase values to the input wavefront calculated from the
    Optical Path Difference (OPD). The OPDs are calculated from the basis
    arrays, and weighted by the coefficients, and converted to phases by the
    wavefront methods.

    Attributes
    ----------
    basis: Array, meters
        Arrays holding the pre-calculated basis vectors.
    coefficients: Array
        The Array of coefficients to be applied to each basis vector.
    name : str
        The name of the layer, which is used to index the layers dictionary.
    """
    basis        : Array
    coefficients : Array


    def __init__(self         : OpticalLayer,
                 basis        : Array,
                 coefficients : Array = None,
                 name         : str = 'ApplyBasisOPD') -> OpticalLayer:
        """
        Constructor for the ApplyBasisOPD class.

        Parameters
        ----------
        basis : Array, meters
            The Array of basis polynomials. This should be a 3 dimensional Array
            with the first dimension being the number of basis vectors, and the
            last two dimensions being equal to the wavefront shape at the time
            of application to the wavefront.
        coefficients : Array = None
            The coefficients by which to weight the basis vectors. This must
            have the same length as the first dimension of the basis Array. If
            None is supplied an Array of zeros is constructed.
        name : str = 'ApplyBasisOPD'
            The name of the layer, which is used to index the layers dictionary.
        """
        super().__init__(name)
        self.basis        = np.asarray(basis, dtype=float)
        self.coefficients = np.zeros(basis.shape[0]) if coefficients is None \
                            else np.asarray(coefficients, dtype=float)

        # Input checks
        assert self.basis.ndim == 3, \
        ("basis must be a 3 dimensional array, ie (nterms, npixels, npixels).")
        assert self.coefficients.ndim == 1 and \
        self.coefficients.shape[0] == self.basis.shape[0], \
        ("coefficients must be a 1 dimensional array with length equal to the "
        "First dimension of the basis array.")


    def get_opd(self : OpticalLayer) -> Array:
        """
        A function to calculate the total OPD from the basis vector and the
        coefficients.

        Returns
        -------
        OPD : Array, meters
            The total OPD calulated from the basis vectors and coefficients.
        """
        return np.dot(self.basis.T, self.coefficients)


    def __call__(self : OpticalLayer, wavefront : Wavefront) -> Wavefront:

        """
        Calculate and apply the appropriate phase shift to the wavefront.

        Parameters
        ----------
        wavefront : Wavefront
            The wavefront to operate on.

        Returns
        -------
        wavefront : Wavefront
            The wavefront with the appropriate phase applied.
        """
        return wavefront.add_opd(self.get_opd())
    

    def summary(self            : OpticalLayer, 
                angular_units   : str = 'radians', 
                cartesian_units : str = 'meters', 
                sigfigs         : int = 4) -> str:
        """
        Returns a summary of the class.

        Parameters
        ----------
        angular_units : str = 'radians'
            The angular units to use in the summary. Options are 'radians', 
            'degrees', 'arcseconds' and 'arcminutes'.
        cartesian_units : str = 'meters'
            The cartesian units to use in the summary. Options are 'meters',
            'millimeters' and 'microns'.
        sigfigs : int = 4
            The number of significant figures to use in the summary.

        Returns
        -------
        summary : str
            A summary of the class.
        """
        return (f"{self.name}: Applies an Optical Path Difference (OPD) to the "
            f"wavefront calculated from the basis vectors and coefficients.")


class AddPhase(OpticalLayer):
    """
    Adds an array of phase values to the wavefront.

    Attributes
    ----------
    phase: Array, radians
        The Array of phase values to be applied to the input wavefront.
    name : str
        The name of the layer, which is used to index the layers dictionary.
    """
    phase : Array


    def __init__(self  : OpticalLayer,
                 phase : Array,
                 name  : str = 'AddPhase') -> OpticalLayer:
        """
        Constructor for the AddPhase class.

        Parameters
        ----------
        phase : Array, radians
            Array of phase values to be applied to the input wavefront. This
            must a 0, 2 or 3 dimensional array with equal to that of the 
            wavefront at time of aplication.
        name : str = 'AddPhase'
            The name of the layer, which is used to index the layers dictionary.
        """
        super().__init__(name)
        self.phase = np.asarray(phase, dtype=float)

        # Input checks
        assert self.phase.ndim in (0, 2, 3), ("phase must be either a scalar "
        "array, or a 2 or 3 dimensional array.")


    def __call__(self : OpticalLayer, wavefront : Wavefront) -> Wavefront:
        """
        Adds the phase to the wavefront.

        Parameters
        ----------
        wavefront : Wavefront
            The wavefront to operate on.

        Returns
        -------
        wavefront : Wavefront
            The wavefront with the phase added.
        """
        return wavefront.add_phase(self.phase)
    

    def summary(self            : OpticalLayer, 
                angular_units   : str = 'radians', 
                cartesian_units : str = 'meters', 
                sigfigs         : int = 4) -> str:
        """
        Returns a summary of the class.

        Parameters
        ----------
        angular_units : str = 'radians'
            The angular units to use in the summary. Options are 'radians', 
            'degrees', 'arcseconds' and 'arcminutes'.
        cartesian_units : str = 'meters'
            The cartesian units to use in the summary. Options are 'meters',
            'millimeters' and 'microns'.
        sigfigs : int = 4
            The number of significant figures to use in the summary.

        Returns
        -------
        summary : str
            A summary of the class.
        """
        return f"{self.name}: Add an array of phase values to the wavefront."


class AddOPD(OpticalLayer):
    """
    Adds an Optical Path Difference (OPD) to the wavefront.

    Attributes
    ----------
    opd : Array, meters
        Array of OPD values to be applied to the input wavefront.
    name : str
        The name of the layer, which is used to index the layers dictionary.
    """
    opd : Array


    def __init__(self : OpticalLayer,
                 opd  : Array,
                 name : str = 'AddOPD') -> OpticalLayer:
        """
        Constructor for the ApplyOPD class.

        Parameters
        ----------
        opd : float, meters
            The Array of OPDs to be applied to the input wavefront. This must
            a 0, 2 or 3 dimensional array with equal to that of the wavefront
            at time of aplication.
        name : str = 'AddOPD'
            The name of the layer, which is used to index the layers dictionary.
        """
        super().__init__(name)
        self.opd = np.asarray(opd, dtype=float)

        # Input checks
        assert self.opd.ndim in (0, 2, 3), ("opd must be either a scalar "
        "array, or a 2 or 3 dimensional array.")


    def __call__(self : OpticalLayer, wavefront : Wavefront) -> Wavefront:
        """
        Apply the OPD array to the input wavefront.

        Parameters
        ----------
        wavefront : Wavefront
            The wavefront to operate on.

        Returns
        -------
        wavefront : Wavefront
            The wavefront with the OPD added.
        """
        return wavefront.add_opd(self.opd)
    

    def summary(self            : OpticalLayer, 
                angular_units   : str = 'radians', 
                cartesian_units : str = 'meters', 
                sigfigs         : int = 4) -> str:
        """
        Returns a summary of the class.

        Parameters
        ----------
        angular_units : str = 'radians'
            The angular units to use in the summary. Options are 'radians', 
            'degrees', 'arcseconds' and 'arcminutes'.
        cartesian_units : str = 'meters'
            The cartesian units to use in the summary. Options are 'meters',
            'millimeters' and 'microns'.
        sigfigs : int = 4
            The number of significant figures to use in the summary.

        Returns
        -------
        summary : str
            A summary of the class.
        """
        return (f"{self.name}: Add an array of Optical Path Differences (OPD) "
                "to the wavefront.")


class TransmissiveOptic(OpticalLayer):
    """
    Represents an arbitrary transmissive optic.

    Note this class does not normalise the 'transmission' between 0 and 1, but
    simply multiplies the wavefront amplitude by the transmision array.

    Attributes
    ----------
    transmission : Array
        An array representing the transmission of the optic.
    name : str
        The name of the layer, which is used to index the layers dictionary.
    """
    transmission: Array


    def __init__(self         : OpticalLayer,
                 transmission : Array,
                 name         : str = 'TransmissiveOptic') -> OpticalLayer:
        """
        Constructor for the TransmissiveOptic class.

        Parameters
        ----------
        transmission : Array
            The array representing the transmission of the aperture. This must
            a 0, 2 or 3 dimensional array with equal to that of the wavefront
            at time of aplication.
        name : str = 'TransmissiveOptic'
            The name of the layer, which is used to index the layers dictionary.
        """
        super().__init__(name)
        self.transmission = np.asarray(transmission, dtype=float)

        # Input checks
        assert self.transmission.ndim in (0, 2, 3), ("transmission must be "
        "either a scalar array, or a 2 or 3 dimensional array.")


    def __call__(self : OpticalLayer, wavefront : Wavefront) -> Wavefront:
        """
        Applies the tranmission of the optical to the wavefront.

        Parameters
        ----------
        wavefront : Wavefront
            The wavefront to operate on.

        Returns
        -------
        wavefront : Wavefront
            The wavefront with the tranmission applied.
        """
        return wavefront.multiply_amplitude(self.transmission)
    

    def summary(self            : OpticalLayer, 
                angular_units   : str = 'radians', 
                cartesian_units : str = 'meters', 
                sigfigs         : int = 4) -> str:
        """
        Returns a summary of the class.

        Parameters
        ----------
        angular_units : str = 'radians'
            The angular units to use in the summary. Options are 'radians', 
            'degrees', 'arcseconds' and 'arcminutes'.
        cartesian_units : str = 'meters'
            The cartesian units to use in the summary. Options are 'meters',
            'millimeters' and 'microns'.
        sigfigs : int = 4
            The number of significant figures to use in the summary.

        Returns
        -------
        summary : str
            A summary of the class.
        """
        return (f"{self.name}: Applies an array of tranmission values to the "
                "Wavefront.")


class ApplyBasisCLIMB(OpticalLayer):
    """
    Adds an array of binary phase values to the input wavefront from a set of
    continuous basis vectors. This uses the CLIMB algorithm in order to
    generate the binary values in a continous manner as described in the
    paper Wong et al. 2021. The basis vectors are taken as an Optical Path
    Difference (OPD), and applied to the phase of the wavefront. The ideal
    wavelength parameter described the wavelength that will have a perfect
    anti-phase relationship given by the Optical Path Difference.

    Note: Many of the methods in the class still need doccumentation.
    Note: This currently only outputs 256 pixel arrays and uses a 3x oversample,
    therefore requiring a 768 pixel basis array.

    Attributes
    ----------
    basis: Array
        Arrays holding the continous pre-calculated basis vectors.
    coefficients: Array
        The Array of coefficients to be applied to each basis vector.
    ideal_wavelength : Array
        The target wavelength at which a perfect anti-phase relationship is
        applied via the OPD.
    name : str
        The name of the layer, which is used to index the layers dictionary.
    """
    basis            : Array
    coefficients     : Array
    ideal_wavelength : Array


    def __init__(self             : OpticalLayer,
                 basis            : Array,
                 ideal_wavelength : Array,
                 coefficients     : Array = None,
                 name             : str   = 'ApplyBasisCLIMB') -> OpticalLayer:
        """
        Constructor for the ApplyBasisCLIMB class.

        Parameters
        ----------
        basis : Array
            Arrays holding the continous pre-calculated basis vectors. This must
            be a 3d array of shape (nterms, npixels, npixels), with the final
            two dimensions matching that of the wavefront at time of
            application. This is currently required to be a nx768x768 shaped
            array. 
        ideal_wavelength : Array
            The target wavelength at which a perfect anti-phase relationship is
            applied via the OPD.
        coefficients : Array = None
            The Array of coefficients to be applied to each basis vector. This
            must be a one dimensional array with leading dimension equal to the
            leading dimension of the basis vectors. Default is None which
            initialises an array of zeros.
        name : str = 'ApplyBasisCLIMB'
            The name of the layer, which is used to index the layers dictionary.
        """
        super().__init__(name)
        self.basis            = np.asarray(basis, dtype=float)
        self.ideal_wavelength = np.asarray(ideal_wavelength, dtype=float)
        self.coefficients     = np.array(coefficients).astype(float) \
                    if coefficients is not None else np.zeros(len(self.basis))

        # Inputs checks
        assert self.basis.ndim == 3, \
        ("basis must be a 3 dimensional array, ie (nterms, npixels, npixels).")
        assert self.basis.shape[-1] == 768, \
        ("Basis must have shape (n, 768, 768).")
        assert self.coefficients.ndim == 1 and \
        self.coefficients.shape[0] == self.basis.shape[0], \
        ("coefficients must be a 1 dimensional array with length equal to the "
        "First dimension of the basis array.")
        assert self.ideal_wavelength.ndim == 0, ("ideal_wavelength must be a "
                                                 "scalar array.")


    def __call__(self : OpticalLayer, wavefront : Wavefront) -> Wavefront:
        """
        Generates and applies the binary OPD array to the wavefront in a
        differentiable manner.

        Parameters
        ----------
        wavefront : Wavefront
            The wavefront to operate on.

        Returns
        -------
        wavefront : Wavefront
            The wavefront with the binary OPD applied.
        """
        latent = self.get_opd(self.basis, self.coefficients)
        binary_phase = np.pi*self.CLIMB(latent, ppsz=wavefront.npixels)
        opd = self.phase_to_opd(binary_phase, self.ideal_wavelength)
        return wavefront.add_opd(opd)


    def opd_to_phase(self, opd, wavel):
        return 2*np.pi*opd/wavel


    def phase_to_opd(self, phase, wavel):
        return phase*wavel/(2*np.pi)


    def get_opd(self, basis, coefficients):
        return np.dot(basis.T, coefficients)


    def get_total_opd(self):
        return self.get_opd(self.basis, self.coefficients)


    def get_binary_phase(self):
        latent = self.get_opd(self.basis, self.coefficients)
        binary_phase = np.pi*self.CLIMB(latent)
        return binary_phase


    def lsq_params(self, img):
        xx, yy = np.meshgrid(np.linspace(0,1,img.shape[0]),
                             np.linspace(0,1,img.shape[1]))
        A = np.vstack([xx.ravel(), yy.ravel(), np.ones_like(xx).ravel()]).T
        matrix = np.linalg.inv(np.dot(A.T,A)).dot(A.T)
        return matrix, xx, yy, A


    def lsq(self, img):
        matrix, _, _, _ = self.lsq_params(img)
        return np.dot(matrix,img.ravel())


    def area(self, img, epsilon = 1e-15):
        a,b,c = self.lsq(img)
        a = np.where(a==0,epsilon,a)
        b = np.where(b==0,epsilon,b)
        c = np.where(c==0,epsilon,c)
        x1 = (-b-c)/(a) # don't divide by zero
        x2 = -c/(a) # don't divide by zero
        x1, x2 = np.min(np.array([x1,x2])), np.max(np.array([x1,x2]))
        x1, x2 = np.max(np.array([x1,0])), np.min(np.array([x2,1]))

        dummy = x1 + (-c/b)*x2-(0.5*a/b)*x2**2 - (-c/b)*x1+(0.5*a/b)*x1**2

        # Set the regions where there is a defined gradient
        dummy = np.where(dummy>=0.5,dummy,1-dummy)

        # Colour in regions
        dummy = np.where(np.mean(img)>=0,dummy,1-dummy)

        # rescale between 0 and 1?
        dummy = np.where(np.all(img>0),1,dummy)
        dummy = np.where(np.all(img<=0),0,dummy)

        # undecided region
        dummy = np.where(np.any(img==0),np.mean(dummy>0),dummy)

        # rescale between 0 and 1
        dummy = np.clip(dummy, 0, 1)

        return dummy

    def CLIMB(self, wf, ppsz = 256):
        psz = ppsz * 3
        dummy = np.array(np.split(wf, ppsz))
        dummy = np.array(np.split(np.array(dummy), ppsz, axis = 2))
        subarray = dummy[:,:,0,0]

        flat = dummy.reshape(-1, 3, 3)
        vmap_mask = vmap(self.area, in_axes=(0))

        soft_bin = vmap_mask(flat).reshape(ppsz, ppsz)

        return soft_bin
    

    def summary(self            : OpticalLayer, 
                angular_units   : str = 'radians', 
                cartesian_units : str = 'meters', 
                sigfigs         : int = 4) -> str:
        """
        Returns a summary of the class.

        Parameters
        ----------
        angular_units : str = 'radians'
            The angular units to use in the summary. Options are 'radians', 
            'degrees', 'arcseconds' and 'arcminutes'.
        cartesian_units : str = 'meters'
            The cartesian units to use in the summary. Options are 'meters',
            'millimeters' and 'microns'.
        sigfigs : int = 4
            The number of significant figures to use in the summary.

        Returns
        -------
        summary : str
            A summary of the class.
        """
        return (f"{self.name}: Applies a binary OPD to the Wavefront using the "
                "CLIMB algorithm.")


class Rotate(OpticalLayer):
    """
    Applies a rotation to the wavefront using interpolation methods.

    Attributes
    ----------
    angle : Array, radians
        The angle by which to rotate the wavefront in the clockwise direction.
    real_imaginary : bool
        Should the rotation be performed on the amplitude and phase array
        or the real and imaginary arrays.
    fourier : bool
        Should the rotation be done using fourier methods or interpolation.
    order : int = 1
        The order of the interpolation to use. Only applies if fourier is
        False. Must be 0, 1, or 3.
    padding : int
        The amount of padding to use if the fourier method is used.
    name : str
        The name of the layer, which is used to index the layers dictionary.
    """
    angle          : Array
    real_imaginary : bool
    fourier        : bool
    order          : int
    padding        : int


    def __init__(self           : OpticalLayer,
                 angle          : Array,
                 real_imaginary : bool = False,
                 fourier        : bool = False,
                 order          : int  = 1,
                 padding        : int  = None,
                 name           : str  = 'Rotate') -> OpticalLayer:
        """
        Constructor for the Rotate class.

        Parameters
        ----------
        angle: float, radians
            The angle by which to rotate the wavefront in the clockwise 
            direction.
        real_imaginary : bool = False
            Should the rotation be performed on the amplitude and phase array
            or the real and imaginary arrays.
        fourier : bool = False
            Should the fourier rotation method be used (True), or regular
            interpolation method be used (False).
        order : int = 1
            The order of the interpolation to use. Only applies if fourier is
            False. Must be 0, 1, or 3.
        padding : int = None
            The amount of fourier padding to use. Only applies if fourier is
            True.
        name : str = 'Rotate'
            The name of the layer, which is used to index the layers dictionary.
        """
        super().__init__(name)
        self.angle          = np.asarray(angle, dtype=float)
        self.real_imaginary = bool(real_imaginary)
        if order not in (0, 1, 3):
            raise ValueError("Order must be 0, 1, or 3.")
        self.order = int(order)
        self.fourier        = bool(fourier)
        self.padding = padding if padding is None else int(padding)
        assert self.angle.ndim == 0, ("angle must be scalar array.")


    def __call__(self : OpticalLayer, wavefront : Wavefront) -> Wavefront:
        """
        Applies the rotation to a wavefront.

        Parameters
        ----------
        wavefront : Wavefront
            The wavefront to operate on.

        Returns
        -------
        wavefront : Wavefront
            The rotated wavefront.
        """
        args = [self.angle, self.real_imaginary, self.fourier, self.order]
        args += [self.padding] if self.padding is not None else []
        return wavefront.rotate(*args)


    def summary(self            : OpticalLayer, 
                angular_units   : str = 'radians', 
                cartesian_units : str = 'meters', 
                sigfigs         : int = 4) -> str:
        """
        Returns a summary of the class.

        Parameters
        ----------
        angular_units : str = 'radians'
            The angular units to use in the summary. Options are 'radians', 
            'degrees', 'arcseconds' and 'arcminutes'.
        cartesian_units : str = 'meters'
            The cartesian units to use in the summary. Options are 'meters',
            'millimeters' and 'microns'.
        sigfigs : int = 4
            The number of significant figures to use in the summary.

        Returns
        -------
        summary : str
            A summary of the class.
        """
        angle = convert_angular(self.angle, 'radians', angular_units)
        
        if self.fourier:
            method = f"a Fourier method with padding of {self.padding}"
        else:
            method = "an Interpolation method of order 1"

        if self.real_imaginary:
            wf_type = "real and imaginary arrays"
        else:
            wf_type = "amplitude and phase arrays"

        return (f"{self.name}: Applies a {angle:.{sigfigs}} {angular_units} "
                f"rotation to the wavefront {wf_type} using {method}.")


import jax.numpy as np
import dLux


__all__ = ["simple_optical_system", "toliman"]


Array = np.ndarray


def simple_optical_system(aperture_diameter         : Array,
                          wavefront_npixels         : int,
                          detector_npixels          : int,
                          detector_pixel_size       : Array,
                          angular                   : bool  = True,
                          focal_length              : Array = None,
                          secondary_mirror_diameter : Array = None,
                          nzernike                  : int   = None,
                          zernike_coefficients      : Array = None,
                          extra_layers              : list  = None,
                          return_layers             : bool  = False):
    """
    Constucts a simple Fourier optical system.

    Parameters
    ----------
    aperture_diameter : Array, meters
        The diameter of the optical system aperture.
    wavefront_npixels : int
        The number of pixel used to represent the wavefront.
    detector_npixels : int
        The number of pixel of the detector
    detector_pixel_size : Array, arcseconds/pixel or meters/pixel
        The size of the detector pixels. Taken in units of arcseconds per pixel
        if anuglar == True, else units are taken in meters per pixel.
    angular : bool = True
        Whether to use angular (radians) or cartesian (meters) units.
    focal_length : Array = None
        The focal length of the optical system. This paramter is only used if
        angular == False.
    secondary_mirror_diameter : Array = None
        The diameter of the secondary mirror obscuration.
    nzernike : int = None
        The number of zernike terms to use. Ignore piston tip tilt.
    zernike_coefficients : Array = None
        The values of the zernike coefficients. Only used if nzerike == None.
    extra_layers : list = None
        The extra layers to add to the optical system.
    return_layers : bool = False
        Should the function return the layers, or an Optics class.

    Returns
    -------
    optics : OpticalSystem
        The optical system with the optical layers loaded.
    """
    # Inputs checks
    aperture_diameter = np.asarray(aperture_diameter, dtype=float)
    assert aperture_diameter.ndim == 0, ("aperture_diameter must be scalar.")

    assert isinstance(wavefront_npixels, int), \
    ("wavefront_npixels must be an integer.")

    assert isinstance(detector_npixels, int), \
    ("detector_npixels must be an integer.")
    detector_pixel_size = np.asarray(detector_pixel_size, dtype=float)

    assert detector_pixel_size.ndim == 0, \
    ("detector_pixel_size must be scalar.")

    assert isinstance(angular, bool), ("angular must be a boolean.")

    if not angular:
        focal_length = np.asarray(focal_length, dtype=float)
        assert focal_length.ndim == 0, ("focal_length must be scalar.")

    if secondary_mirror_diameter is not None:
        secondary_mirror_diameter = np.asarray(secondary_mirror_diameter, \
                                               dtype=float)
        assert secondary_mirror_diameter.ndim == 0, \
        ("focal_length must be scalar.")

    if nzernike is not None:
        assert isinstance(nzernike, int), ("nzernike must an integer.")

    if zernike_coefficients is not None:
        zernike_coefficients = np.asarray(zernike_coefficients, dtype=float)
        assert zernike_coefficients.ndim == 1, \
        ("zernike_coefficients must a one dimensional array.")
        assert len(zernike_coefficients) == nzernike, \
        ("The lenght of zernike_coefficients must be equal to nzerike.")

    if extra_layers is not None:
        assert isinstance(extra_layers, list), ("extra_layers must be a list.")
        for layer in extra_layers:
            assert isinstance(layer, dLux.optics.OpticalLayer), \
            ("Each item in extra_layers must be a dLux OpticalLayer.")

    assert isinstance(return_layers, bool), "return_layers must be a boolean."

    # Create wavefront
    if angular:
        layers = [dLux.optics.CreateWavefront(wavefront_npixels,
                                              aperture_diameter,
                                              wavefront_type="Angular")]
    else:
        layers = [dLux.optics.CreateWavefront(wavefront_npixels,
                                              aperture_diameter)]

    # Construct aperture
    if secondary_mirror_diameter is not None:
        secondary_ratio = secondary_mirror_diameter/aperture_diameter
    else:
        secondary_ratio = 0.
    
    if nzernike is not None:
        zernikes = np.arange(3, nzernike + 3)
        if zernike_coefficients is not None:
            coeffs = zernike_coefficients
        else:
            coeffs = np.zeros(nzernike)
    else:
        zernikes = None
        coeffs = None
    
    layers +=[dLux.ApertureFactory(wavefront_npixels, 
                secondary_ratio=secondary_ratio, zernikes=zernikes, 
                coefficients=coeffs)]

    # Extra Layers
    if extra_layers is not None:
        for layer in extra_layers:
            layers.append(layer)

    # Normalised wavefront
    layers += [dLux.NormaliseWavefront()]

    # Propagator
    if angular:
        layers += [dLux.AngularMFT(detector_npixels,
                        dLux.utils.arcseconds_to_radians(detector_pixel_size))]
    else:
        layers += [dLux.CartesianMFT(detector_npixels, detector_pixel_size,
                                   focal_length)]

    # Return optics or layers
    return layers if return_layers else dLux.core.Optics(layers)


def toliman(wavefront_npixels         : int,
            detector_npixels          : int,
            aperture_diameter         : Array = 0.13,
            secondary_mirror_diameter : Array = 0.032,
            detector_pixel_size       : Array = 0.375,
            angular                   : bool  = True,
            focal_length              : Array = None,
            nzernike                  : int   = None,
            zernike_coefficients      : Array = None,
            extra_layers              : list  = None,
            return_layers             : bool  = False):
    """
    Gets a simple Toliman optical system by calling the simple_optical_system
    function with pre-loaded values.

    Parameters
    ----------
    wavefront_npixels : int
        The number of pixel used to represent the wavefront.
    detector_npixels : int
        The number of pixel of the detector
    aperture_diameter : Array, meters = 0.13
        The diameter of the optical system aperture.
    secondary_mirror_diameter : Array = 0.032
        The diameter of the secondary mirror obscuration.
    detector_pixel_size : Array, arcseconds/pixel or meters/pixel = 0.375
        The size of the detector pixels. Taken in units of arcseconds per pixel
        if anuglar == True, else units are taken in meters per pixel. The
        physical detector pixels are 2.4um
    angular : bool = True
        Whether to use angular (radians) or cartesian (meters) units.
    focal_length : Array = None
        The focal length of the optical system. This paramter is only used if
        angular == False. The EFL is 1.32m.
    nzernike : int = None
        The number of zernike terms to use. Ignore piston tip tilt.
    zernike_coefficients : Array = None
        The values of the zernike coefficients. Only used if nzerike == None.
    extra_layers : list = None
        The extra layers to add to the optical system.
    return_layers : bool = False
        Should the function return the layers, or an Optics class.

    Returns
    -------
    optics : OpticalSystem
        The optical system with the optical layers loaded.
    """
    return simple_optical_system(
                        aperture_diameter,
                        wavefront_npixels,
                        detector_npixels,
                        detector_pixel_size=detector_pixel_size,
                        angular=angular,
                        focal_length=focal_length,
                        secondary_mirror_diameter=secondary_mirror_diameter,
                        nzernike=nzernike,
                        zernike_coefficients=zernike_coefficients,
                        extra_layers=extra_layers,
                        return_layers=return_layers)


import dLux 
import jax
import jax.numpy as np
import typing

jax.config.update("jax_debug_nans", True)

Aperture = typing.TypeVar("Aperture")
Array = typing.TypeVar("Array")
Spider = typing.TypeVar("Spider")


class TestAperturesCommonInterfaces():
    """
    For each type of aperture that has common properties, test it
    """
    def _assert_valid_hard_aperture(aperture, msg=''):
        assert ((aperture == 1.) | (aperture == 0.)).all(), msg
        
    def _assert_valid_soft_aperture(aperture, msg=''):
        assert (aperture <= 1. + np.finfo(np.float32).resolution).all(), msg
        assert (aperture >= 0. - np.finfo(np.float32).resolution).all(), msg
        # there should also exist something in bounds (assuming edges are within coordinates)
        assert np.logical_and(aperture > 0., aperture < 1.).any(), msg
        
    # input is all the fixtures that need testing
    def test_all_apertures(self, create_square_aperture : callable,
                           create_rectangular_aperture : callable,
                           create_circular_aperture : callable,
                           create_hexagonal_aperture : callable,
                           create_annular_aperture : callable):
        
        constructors = [create_square_aperture,
                        create_rectangular_aperture,
                        create_circular_aperture,
                        create_hexagonal_aperture,
                        create_annular_aperture]
        
        # TODO might need to add error message for when this fails but it checks things very easily
        for ctor in constructors:
            self._test_single_aperture_class(ctor)
    
    def _test_single_aperture_class(self, aperture_fixture):
        coordinates = dLux.utils.get_pixel_positions((16, 16), (2./16, 2./16))
        
        x_offset = 1.
        y_offset = 1.
        centres = [np.array([0., 0.]),
                   np.array([x_offset, 0.]),
                   np.array([0., y_offset]),
                   np.array([x_offset, y_offset])]
        
        rotations = [0, np.pi/3., -np.pi/3.5, np.pi/2.]
        not_rotatable_apertures = (dLux.apertures.CircularAperture,
                                   dLux.apertures.AnnularAperture)
        
        base_kwargs = {"centre" : None,
                       "softening" : None,
                       "occulting" : None
                       }

        for centre in centres:
            for rotation in rotations:
                for softening in [0., 1.]:
                    for occulting in [True, False]:
                        actual_kwargs = base_kwargs
                        actual_kwargs["centre"] = centre
                        actual_kwargs["softening"] = softening
                        actual_kwargs["occulting"] = occulting
                        
                        if not isinstance(aperture_fixture(), not_rotatable_apertures):
                            actual_kwargs["rotation"] = rotation
                            
                        
                        aperture = aperture_fixture(**actual_kwargs)._aperture(coordinates)
                        
                        msg = f'{actual_kwargs}, on ctor {aperture_fixture}'
                        
                        if softening:
                            TestAperturesCommonInterfaces._assert_valid_soft_aperture(aperture, msg)
                        else:
                            TestAperturesCommonInterfaces._assert_valid_hard_aperture(aperture, msg)


class TestUniformSpider(object):
    """
    Contains the unit tests for the `UniformSpider` class.
    """


    def test_constructor(self, create_uniform_spider: callable) -> None:
        """
        Tests that the state is correctly initialised. 
        """
        # Test functioning
        spider = create_uniform_spider()

    
    def test_range_hard(self, create_uniform_spider: callable) -> None:
        """
        Checks that the apertures fall into the correct range.
        """
        npix = 16
        width = 2.
        pixel_scale = width / npix
        coordinates = dLux.utils.get_pixel_positions((npix,npix), 
                                                    (pixel_scale, pixel_scale))

        # Case Translated 
        spider = create_uniform_spider(centre=[1., 1.], softening=0.)
        aperture = spider._aperture(coordinates)
        assert ((aperture == 1.) | (aperture == 0.)).all()

        # Case Rotated 
        spider = create_uniform_spider(rotation=np.pi/4., softening=0.)
        aperture = spider._aperture(coordinates)
        assert ((aperture == 1.) | (aperture == 0.)).all()

        # Case Strained 
        spider = create_uniform_spider(shear=[.05, .05], softening=0.)
        aperture = spider._aperture(coordinates)
        assert ((aperture == 1.) | (aperture == 0.)).all()

        # Case Compression
        spider = create_uniform_spider(compression=[1.05, .95], softening=0.)
        aperture = spider._aperture(coordinates)
        assert ((aperture == 1.) | (aperture == 0.)).all()


    def test_range_soft(self, create_uniform_spider: callable) -> None:
        """
        Checks that the aperture falls into the correct range.
        """
        npix = 16
        width = 2.
        pixel_scale = width / npix
        coordinates = dLux.utils.get_pixel_positions((npix,npix), 
                                                    (pixel_scale, pixel_scale))

        # Case Translated 
        spider = create_uniform_spider(centre=[1., 1.], softening=0.)
        aperture = spider._aperture(coordinates)
        assert (aperture <= 1. + np.finfo(np.float32).resolution).all()
        assert (aperture >= 0. - np.finfo(np.float32).resolution).all()

        # Case Rotated 
        spider = create_uniform_spider(rotation=np.pi/4., softening=0.)
        aperture = spider._aperture(coordinates)
        assert (aperture <= 1. + np.finfo(np.float32).resolution).all()
        assert (aperture >= 0. - np.finfo(np.float32).resolution).all()

        # Case Strained 
        spider = create_uniform_spider(shear=[.05, .05], softening=0.)
        aperture = spider._aperture(coordinates)
        assert (aperture <= 1. + np.finfo(np.float32).resolution).all()
        assert (aperture >= 0. - np.finfo(np.float32).resolution).all()

        # Case Compression
        spider = create_uniform_spider(compression=[1.05, .95], softening=0.)
        aperture = spider._aperture(coordinates)
        assert (aperture <= 1. + np.finfo(np.float32).resolution).all()
        assert (aperture >= 0. - np.finfo(np.float32).resolution).all()


class TestAberratedAperture(object):
    """
    Checks that the aberrated aperture is functional. It does not
    test whether the aberrated aperture is correct.
    """


    def test_constructor(self, create_aberrated_aperture: callable) -> None:
        """
        Tests that it is possible to instantiate an AberratedAperture.
        Does not test if the AberratedAperture is correct.
        """
        # TODO: Make sure that the class asserts that the coeffs and 
        # the noll indexes have the same length.
        create_aberrated_aperture()


    def test_on_aperture(self: object, 
            create_aberrated_aperture: callable,
            create_circular_aperture: callable) -> None:
        """
        Tests that the basis functions are evaluated atop the aperture.
        Applies mutliple different permutations.
        """
        jax.config.update("jax_debug_nans", True)
        
        npix = 16
        width = 2.
        pixel_scale = width / npix
        coordinates = dLux.utils.get_pixel_positions((npix,npix), 
                                                    (pixel_scale, pixel_scale))

        ap = create_circular_aperture()

        aber_ap = create_aberrated_aperture(aperture=ap)._basis(coordinates)
        ap = ap._aperture(coordinates)

        abers = np.where(ap == 0., aber_ap, 0.)
        assert (abers == 0.).all()




import jax.numpy as np

__all__ = ["convert_cartesian", "convert_angular",
           "radians_to_arcseconds", "radians_to_degrees", 
           "radians_to_arcminutes", "degrees_to_radians", 
           "degrees_to_arcminutes", "degrees_to_arcseconds", 
           "arcminutes_to_radians", "arcminutes_to_degrees", 
           "arcminutes_to_arcseconds", "arcseconds_to_radians", 
           "arcseconds_to_degrees", "arcseconds_to_arcminutes",
           "r2s", "r2d", "r2m", "d2r", "d2m", "d2s", 
           "m2r", "m2d", "m2s", "s2r", "s2d", "s2m"]

Array = np.ndarray


# General conversion classes:
def convert_cartesian(values : Array, 
                      input  : str = 'meters', 
                      output : str = 'meters') -> Array:
    """
    Converts the input values from one unit to another.

    Parameters
    ----------
    values : Array
        The input values to be converted.
    input : str = 'meters'
        The input units. Must be one of 'meters', 'millimeters', or 'microns'.
    output : str = 'meters'
        The output units. Must be one of 'meters', 'millimeters', or 'microns'.
    
    Returns
    -------
    values : Array
        The input values converted into the output units.
    """
    if input not in ('meters', 'millimeters', 'microns'):
        raise ValueError("input must be 'meters', 'millimeters', or 'microns'.")
    if output not in ('meters', 'millimeters', 'microns'):
        raise ValueError("output must be 'meters', 'millimeters', or "
                         "'microns'.")
    
    if input == output:
        factor = 1
    elif input == 'meters':
        if output == 'millimeters':
            factor = 1e-3
        elif output == 'microns':
            factor = 1e-6
    elif input == 'millimeter':
        if output == 'meters':
            factor = 1e3
        elif output == 'microns':
            factor = 1e-3
    elif input == 'microns':
        if output == 'meters':
            factor = 1e6
        elif output == 'millimeters':
            factor = 1e3
    return values * factor


def convert_angular(values : Array,
                    input  : str = 'radians',
                    output : str = 'radians') -> Array:
    """
    Converts the input values from one unit to another.

    Parameters
    ----------
    values : Array
        The input values to be converted.
    input : str = 'radians'
        The input units. Must be one of 'radians', 'degrees', 'arcseconds', or
        'arcminutes'.
    output : str = 'radians'
        The output units. Must be one of 'radians', 'degrees', 'arcseconds', or
        'arcminutes'.
    
    Returns
    -------
    values : Array
        The input values converted into the output units.
    """
    if input not in ('radians', 'degrees', 'arcseconds', 'arcminutes'):
        raise ValueError(f"input must be one of 'radians', 'degrees', "
                         f"'arcseconds' or 'arcminutes'.")
    if output not in ('radians', 'degrees', 'arcseconds', 'arcminutes'):
        raise ValueError(f"output must be one of 'radians', 'degrees', "
                         f"'arcseconds' or 'arcminutes'.")

    if input == output:
        factor = 1
    elif input == 'radians':
        if output == 'degrees':
            factor = r2d(1)
        elif output == 'arcminutes':
            factor = r2m(1)
        elif output == 'arcseconds':
            factor = r2s(1)
    elif input == 'degrees':
        if output == 'radians':
            factor = d2r(1)
        elif output == 'arcminutes':
            factor = d2m(1)
        elif output == 'arcseconds':
            factor = d2s(1)
    elif input == 'arcminutes':
        if output == 'radians':
            factor = m2r(1)
        elif output == 'degrees':
            factor = m2d(1)
        elif output == 'arcseconds':
            factor = m2s(1)
    elif input == 'arcseconds':
        if output == 'radians':
            factor = s2r(1)
        elif output == 'degrees':
            factor = s2d(1)
        elif output == 'arcminutes':
            factor = s2m(1)
    return values * factor


# Radians to:
def radians_to_arcseconds(values : Array) -> Array:
    """
    Converts the inputs values from radians to arcseconds.

    Can also be imported as r2s.

    Parameters
    ----------
    values : Array, radians
        The input values in units of radians to be converted into arcseconds.

    Returns
    -------
    values : Array, arcseconds
        The input values converted into arcseconds.
    """
    return values * 3600 * 180 / np.pi


def radians_to_degrees(values : Array) -> Array:
    """
    Converts the inputs values from radians to degrees.

    Can also be imported as r2d.

    Parameters
    ----------
    values : Array, radians
        The input values in units of radians to be converted into degrees.

    Returns
    -------
    values : Array, degrees
        The input values converted into degrees.
    """
    return values * 180 / np.pi


def radians_to_arcminutes(values : Array) -> Array:
    """
    Converts the inputs values from radians to arcminutes.

    Can also be imported as r2m.

    Parameters
    ----------
    values : Array, radians
        The input values in units of radians to be converted into arcminutes.

    Returns
    -------
    values : Array, arcminutes
        The input values converted into arcminutes.
    """
    return values * 60 * 180 / np.pi


# Degrees to:
def degrees_to_radians(values : Array) -> Array:
    """
    Converts the inputs values from degrees to radians.

    Can also be imported as d2r.

    Parameters
    ----------
    values : Array, degrees
        The input values in units of degrees to be converted into radians.

    Returns
    -------
    values : Array, radians
        The input values converted into radians.
    """
    return values * np.pi / 180


def degrees_to_arcminutes(values : Array) -> Array:
    """
    Converts the inputs values from degrees to arcminutes.

    Can also be imported as d2m.

    Parameters
    ----------
    values : Array, degrees
        The input values in units of degrees to be converted into arcminutes.

    Returns
    -------
    values : Array, arcminutes
        The input values converted into arcminutes.
    """
    return values * 60


def degrees_to_arcseconds(values : Array) -> Array:
    """
    Converts the inputs values from degrees to arcseconds.

    Can also be imported as d2s.

    Parameters
    ----------
    values : Array, degrees
        The input values in units of degrees to be converted into arcseconds.

    Returns
    -------
    values : Array, arcseconds
        The input values converted into arcseconds.
    """
    return values * 3600

# Arcminutes to:
def arcminutes_to_radians(values : Array) -> Array:
    """
    Converts the inputs values from arcminutes to radians.

    Can also be imported as m2r.

    Parameters
    ----------
    values : Array, arcminutes
        The input values in units of arcminutes to be converted into radians.

    Returns
    -------
    values : Array, radians
        The input values converted into radians.
    """
    return values * np.pi / (60 * 180)


def arcminutes_to_degrees(values : Array) -> Array:
    """
    Converts the inputs values from arcminutes to degrees.

    Can also be imported as m2d.

    Parameters
    ----------
    values : Array, arcminutes
        The input values in units of arcminutes to be converted into degrees.

    Returns
    -------
    values : Array, degrees
        The input values converted into degrees.
    """
    return values / 60


def arcminutes_to_arcseconds(values : Array) -> Array:
    """
    Converts the inputs values from arcminutes to arcseconds.

    Can also be imported as m2s.

    Parameters
    ----------
    values : Array, arcminutes
        The input values in units of arcminutes to be converted into arcseconds.

    Returns
    -------
    values : Array, arcseconds
        The input values converted into arcseconds.
    """
    return values * 60


# Arcseconds to:
def arcseconds_to_radians(values : Array) -> Array:
    """
    Converts the inputs values from arcseconds to radians.

    Can also be imported as s2r.

    Parameters
    ----------
    values : Array, arcseconds
        The input values in units of arcseconds to be converted into radians.

    Returns
    -------
    values : Array, radians
        The input values converted into radians.
    """
    return values * np.pi / (3600 * 180)


def arcseconds_to_degrees(values : Array) -> Array:
    """
    Converts the inputs values from arcseconds to degrees.

    Can also be imported as s2d.

    Parameters
    ----------
    values : Array, arcseconds
        The input values in units of arcseconds to be converted into degrees.

    Returns
    -------
    values : Array, degrees
        The input values converted into degrees.
    """
    return values / 3600


def arcseconds_to_arcminutes(values : Array) -> Array:
    """
    Converts the inputs values from arcseconds to arcminutes.

    Can also be imported as s2m.

    Parameters
    ----------
    values : Array, arcseconds
        The input values in units of arcseconds to be converted into arcminutes.

    Returns
    -------
    values : Array, arcminutes
        The input values converted into arcminutes.
    """
    return values / 60

# Alias to simpler names
r2s = radians_to_arcseconds
r2d = radians_to_degrees
r2m = radians_to_arcminutes
d2r = degrees_to_radians
d2m = degrees_to_arcminutes
d2s = degrees_to_arcseconds
m2r = arcminutes_to_radians
m2d = arcminutes_to_degrees
m2s = arcminutes_to_arcseconds
s2r = arcseconds_to_radians
s2d = arcseconds_to_degrees
s2m = arcseconds_to_arcminutes


from __future__ import annotations
import jax.numpy as np
import pytest
import dLux
from jax import config
config.update("jax_debug_nans", True)

Array = np.ndarray

def test_model(
        create_optics: callable,
        create_detector: callable,
        create_point_source: callable) -> None:
    """
    Test the model function
    """
    optics = create_optics()
    detector = create_detector()
    source = create_point_source()
    sources = [source, source]

    # Test non-optics input
    with pytest.raises(AssertionError):
        dLux.core.model([], [])

    # Test non detector input
    with pytest.raises(AssertionError):
        dLux.core.model(optics, sources, detector=[])

    # Test no source inputs
    with pytest.raises(TypeError):
        dLux.core.model(optics)

    # Test sources with non-source input
    with pytest.raises(AssertionError):
        dLux.core.model(optics, sources=[source, 1.])

    # Test different combinations of inputs, detector
    out = dLux.core.model(optics, sources=sources, detector=detector)
    assert not np.isnan(out).all()
    assert not np.isinf(out).all()

    # Test normalise function for scene
    out = dLux.core.model(optics, sources=sources)
    assert not np.isnan(out).all()
    assert not np.isinf(out).all()

    # Test normalise function for sources
    out = dLux.core.model(optics, sources=sources, normalise=False)
    assert not np.isnan(out).all()
    assert not np.isinf(out).all()

    # Test return_tree with different structures, list
    out = dLux.core.model(optics, sources=sources, return_tree=True)
    out = np.array(out)
    assert not np.isnan(out).all()
    assert not np.isinf(out).all()

    # Test return_tree with different structures, dict
    out = dLux.core.model(optics, sources={"source": source}, return_tree=True)
    out = np.array(list(out.values()))
    assert not np.isnan(out).all()
    assert not np.isinf(out).all()

    # Test return_tree with different structures, source
    out = dLux.core.model(optics, sources=sources, return_tree=True)
    out = np.array(out)
    assert not np.isnan(out).all()
    assert not np.isinf(out).all()

    # Test return_tree with detector input
    out = dLux.core.model(optics, sources=sources, detector=detector, return_tree=True)
    out = np.array(out)
    assert not np.isnan(out).all()
    assert not np.isinf(out).all()


class TestOptics(object):
    """
    Tests the Optics class.
    """


    def test_constructor(self, create_optics: callable) -> None:
        """
        Tests the constructor.
        """
        # Test non-list inputs
        with pytest.raises(AssertionError):
            create_optics(layers={})

        # Test list input with non Optics Layer input
        with pytest.raises(AssertionError):
            create_optics(layers=[10.])


    def test_propagate_mono(self, create_optics: callable) -> None:
        """
        Tests the propagate_mono method.
        """
        osys = create_optics() 

        # Test inputs
        with pytest.raises(AssertionError):
            osys.propagate_mono([1e-6])

        with pytest.raises(AssertionError):
            osys.propagate_mono(1e-6, [0.])

        with pytest.raises(AssertionError):
            osys.propagate_mono(1e-6, weight=[0.])

        # Test propagation
        psf = osys.propagate_mono(4e-6)
        assert not np.isnan(psf).all()
        assert not np.isinf(psf).all()


    def test_propagate(self, create_optics: callable) -> None:
        """
        Tests the propagate method.
        """
        osys = create_optics()

        # Test inputs
        with pytest.raises(AssertionError):
            osys.propagate(1e-6)

        with pytest.raises(AssertionError):
            osys.propagate([[1e-6]])

        with pytest.raises(AssertionError):
            osys.propagate([1e-6], [0.])

        with pytest.raises(AssertionError):
            osys.propagate([1e-6, 2e-6], weights=[0.])

        # Test propagation
        psf = osys.propagate([4e-6, 5e-6])
        assert not np.isnan(psf).all()
        assert not np.isinf(psf).all()


    def test_debug_prop(self, create_optics: callable) -> None:
        """
        Tests the debug_prop method.
        """
        osys = create_optics()

        # Test inputs
        with pytest.raises(AssertionError):
            osys.debug_prop([1e-6])

        with pytest.raises(AssertionError):
            osys.debug_prop(1e-6, [0.])

        with pytest.raises(AssertionError):
            osys.debug_prop(1e-6, weight=[0.])

        # Test propagation
        psf, _, _ = osys.debug_prop(4e-6)
        assert not np.isnan(psf).all()
        assert not np.isinf(psf).all()


    def test_model(self, 
            create_optics: callable, 
            create_point_source: callable) -> None:
        """
        Tests the model method
        """
        osys = create_optics()
        psf = osys.model(sources=create_point_source())
        assert not np.isnan(psf).all()
        assert not np.isinf(psf).all()


class TestDetector(object):
    """
    Tests the Detector class.
    """


    def test_constructor(self, create_detector: callable) -> None:
        """
        Tests the constructor.
        """
        # Test non-list inputs
        with pytest.raises(AssertionError):
            create_detector(layers={})

        # Test list input with non Optics Layer input
        with pytest.raises(AssertionError):
            create_detector(layers=[10.])


    def test_apply_detector(self, create_detector: callable) -> None:
        """
        Tests the apply_detector method.
        """
        detector = create_detector()

        # Test inputs
        with pytest.raises(AssertionError):
            detector.apply_detector([1e-6])

        # Test 1d input
        with pytest.raises(AssertionError):
            detector.apply_detector(np.array([1.]))

        # Test 3d input
        with pytest.raises(AssertionError):
            detector.apply_detector(np.array([[[1.]]]))

        # Test propagation
        image = detector.apply_detector(np.ones((5, 5)))
        assert not np.isnan(image).all()
        assert not np.isinf(image).all()


    def test_debug_apply_detector(self, create_detector: callable) -> None:
        """
        Tests the debug_apply_detector method.
        """
        detector = create_detector()

        # Test inputs
        with pytest.raises(AssertionError):
            detector.debug_apply_detector([1e-6])

        # Test 1d input
        with pytest.raises(AssertionError):
            detector.debug_apply_detector(np.array([1.]))

        # Test 3d input
        with pytest.raises(AssertionError):
            detector.debug_apply_detector(np.array([[[1.]]]))

        # Test propagation
        image, _, _ = detector.debug_apply_detector(np.ones((5, 5)))
        assert not np.isnan(image).all()
        assert not np.isinf(image).all()


    def test_model(self,
            create_detector: callable,
            create_point_source: callable,
            create_optics: callable) -> None:
        """
        Tests the model method
        """
        detector = create_detector()
        source = create_point_source()
        psf = create_optics().model(sources=source)
        psf = detector.model(psf)
        assert not np.isnan(psf).all()
        assert not np.isinf(psf).all()


class TestInstrument(object):
    """
    Tests the Optics class.
    """


    def test_constructor(self, create_instrument: callable) -> None:
        """
        Tests the constructor.
        """
        # Test non optics input
        with pytest.raises(ValueError):
            create_instrument(optics=[])

        # Test non detector input
        with pytest.raises(ValueError):
            create_instrument(detector=[])

        # Test non source input
        with pytest.raises(ValueError):
            create_instrument(sources=[1.])


    def test_normalise(self, create_instrument: callable) -> None:
        """
        Tests the normalise method.
        """
        # Test all sources in the scene are normalised
        instrument = create_instrument()
        normalised_instrument = instrument.normalise()
        for source in normalised_instrument.sources.values():
            assert np.allclose(source.get_weights().sum(), 1.)
            if hasattr(source, 'get_distribution'):
                assert np.allclose(source.get_distribution(), 1.)


    def test_model(self, 
            create_instrument: callable, 
            create_point_source: callable) -> None:
        """
        Tests the model method.
        """
        instrument = create_instrument()
        sources = [create_point_source()]
        source = create_point_source()

        # Test modelling
        psf = instrument.model()
        assert not np.isnan(psf).all()
        assert not np.isinf(psf).all()


class TestFilter(object):
    """
    Tests the Filter class.
    """


    def test_constructor(self, create_filter: callable) -> None:
        """
        Tests the constructor.
        """
        # Test adding filter name
        with pytest.raises(NotImplementedError):
            create_filter(filter_name='Test')

        # Test 2d wavelengths input
        with pytest.raises(AssertionError):
            create_filter(2,2)

        # Test 2d throughput input
        with pytest.raises(AssertionError):
            create_filter(2,2)

        # Test different shape wavelengths and throughput
        with pytest.raises(AssertionError):
            create_filter(4)

        # Test negative wavelengths
        with pytest.raises(AssertionError):
            create_filter([-1, 1])

        # Test negative throughputs
        with pytest.raises(AssertionError):
            create_filter([-1, 1])

        # Test throughputs greater than 1
        with pytest.raises(AssertionError):
            create_filter([0, 1.5])

        # Test reverse order wavelengths
        with pytest.raises(AssertionError):
            create_filter([1, 0.5])


    def test_get_throughput(self, create_filter: callable) -> None:
        """
        Test the get_throughput method.
        """
        filt = create_filter()

        # Test scalar input
        throughput = filt.get_throughput(np.array([1e-6, 2e-6]))
        assert not np.isnan(throughput).all()
        assert not np.isinf(throughput).all()
        assert (throughput >= 0).all()
        assert (throughput <= 1).all()

        # test array input
        throughput = filt.get_throughput(np.array([1e-6, 2e-6]))
        assert not np.isnan(throughput).all()
        assert not np.isinf(throughput).all()
        assert (throughput >= 0).all()
        assert (throughput <= 1).all()

        # Test 1d array inputs
        throughput = filt.get_throughput(1e-6*np.linspace(1, 5, 5))
        assert not np.isnan(throughput).all()
        assert not np.isinf(throughput).all()
        assert (throughput >= 0).all()
        assert (throughput <= 1).all()


from jax import lax


__all__ = ["factorial"]

 
 # TODO: Use lax.cond to make n == 0 a safe case
def factorial(n : float) -> float:
    """
    Calculate n! in a jax friendly way. Note that n == 0 is not a 
    safe case.  
 
    Parameters
    ----------
    n : float
        The value to calculate the factorial of.
 
    Returns
    -------
    n! : float
        The factorial of the value.
    """
    return lax.exp(lax.lgamma(n + 1.))



from __future__ import annotations
import jax.numpy as np
from abc import ABC, abstractmethod
from jax.scipy.signal import convolve
from jax.scipy.stats import norm
from zodiax import Base
from dLux.utils.interpolation import rotate, fourier_rotate
from dLux.utils.units import convert_angular
from dLux.utils.helpers import single_image_plot


__all__ = ["ApplyPixelResponse", "ApplyJitter", "ApplySaturation",
           "AddConstant", "IntegerDownsample", "Rotate"]


Array = np.ndarray


class DetectorLayer(Base, ABC):
    """
    A base Detector layer class to help with type checking throuhgout the rest
    of the software.

    Attributes
    ----------
    name : str
        The name of the layer, which is used to index the layers dictionary.
    """
    name : str


    def __init__(self : DetectorLayer,
                 name : str = 'DetectorLayer') -> DetectorLayer:
        """
        Constructor for the DetectorLayer class.

        Parameters
        ----------
        name : str = 'DetectorLayer'
            The name of the layer, which is used to index the layers dictionary.
        """
        self.name = str(name)


    @abstractmethod
    def __call__(self : DetectorLayer, image : Array) -> Array: # pragma: no cover
        """
        Abstract method for Detector Layers
        """
        return
    

    def summary(self            : DetectorLayer, 
                angular_units   : str = 'radians', 
                cartesian_units : str = 'meters', 
                sigfigs         : int = 4) -> str:
        """
        Returns a summary of the class.

        Parameters
        ----------
        angular_units : str = 'radians'
            The angular units to use in the summary. Options are 'radians', 
            'degrees', 'arcseconds' and 'arcminutes'.
        cartesian_units : str = 'meters'
            The cartesian units to use in the summary. Options are 'meters',
            'millimeters' and 'microns'.
        sigfigs : int = 4
            The number of significant figures to use in the summary.

        Returns
        -------
        summary : str
            A summary of the class.
        """
        return f"{self.name} layer has no summary method yet."
    

    def display(self            : DetectorLayer, 
                image           : Array,
                figsize         : tuple = (5, 4),
                dpi             : int = 120,
                angular_units   : str = 'radians', 
                cartesian_units : str = 'meters', 
                sigfigs         : int = 4) -> None:
        """
        Displays a plot of the image propagating through the layer.

        Parameters
        ----------
        image : Array
            The dummy image to propagate though the detector.
        figsize : tuple = (10, 4)
            The size of the figure to display.
        cmap : str = 'inferno'
            The colour map to use.
        dpi : int = 120
            The resolution of the figure.
        angular_units : str = 'radians'
            The angular units to use in the summary. Options are 'radians', 
            'degrees', 'arcseconds' and 'arcminutes'.
        cartesian_units : str = 'meters'
            The cartesian units to use in the summary. Options are 'meters',
            'millimeters' and 'microns'.
        sigfigs : int = 4
            The number of significant figures to use in the summary.
        """
        single_image_plot(image, figsize=figsize, title="PSF", 
                          cbar_label="Counts", cmap='inferno', dpi=dpi)


class ApplyPixelResponse(DetectorLayer):
    """
    Applies a pixel response array to the the input image, via a multiplication.

    Attributes
    ----------
    pixel_response : Array
        The pixel_response to apply to the input image.
    name : str
        The name of the layer, which is used to index the layers dictionary.
    """
    pixel_response : Array


    def __init__(self           : DetectorLayer,
                 pixel_response : Array,
                 name           : str = 'ApplyPixelResponse') -> DetectorLayer:
        """
        Constructor for the ApplyPixelResponse class.

        Parameters
        ----------
        pixel_response : Array
            The pixel_response to apply to the input image. Must be a 2
            dimensional array equal to size of the image at time of application.
        name : str = 'ApplyPixelResponse'
            The name of the layer, which is used to index the layers dictionary.
        """
        super().__init__(name)
        self.pixel_response = np.asarray(pixel_response, dtype=float)
        assert self.pixel_response.ndim == 2, \
        ("pixel_response must be a 2 dimensional array.")


    def __call__(self : DetectorLayer, image) -> Array:
        """
        Applies the pixel response to the input image, via a multiplication.

        Parameters
        ----------
        image : Array
            The image to apply the pixel_response to.

        Returns
        -------
        image : Array
            The image with the pixel_response applied.
        """
        return image * self.pixel_response
    

    def summary(self            : DetectorLayer, 
                angular_units   : str = 'radians', 
                cartesian_units : str = 'meters', 
                sigfigs         : int = 4) -> str:
        """
        Returns a summary of the class.

        Parameters
        ----------
        angular_units : str = 'radians'
            The angular units to use in the summary. Options are 'radians', 
            'degrees', 'arcseconds' and 'arcminutes'.
        cartesian_units : str = 'meters'
            The cartesian units to use in the summary. Options are 'meters',
            'millimeters' and 'microns'.
        sigfigs : int = 4
            The number of significant figures to use in the summary.

        Returns
        -------
        summary : str
            A summary of the class.
        """
        return f"{self.name}: Applies the pixel response to the image."


class ApplyJitter(DetectorLayer):
    """
    Convolves the image with a gaussian kernel parameterised by the standard
    deviation (sigma).

    Attributes
    ----------
    sigma : Array, pixels
        The standard deviation of the guassian kernel, in units of pixels.
    kernel_size : int
        The size of the convolution kernel to use.
    name : str
        The name of the layer, which is used to index the layers dictionary.
    """
    kernel_size : int
    sigma       : Array


    def __init__(self        : DetectorLayer,
                 sigma       : Array,
                 kernel_size : int = 10,
                 name        : str = 'ApplyJitter') -> DetectorLayer:
        """
        Constructor for the ApplyJitter class.

        Parameters
        ----------
        sigma : Array, pixels
            The standard deviation of the guassian kernel, in units of pixels.
        kernel_size : int = 10
            The size of the convolution kernel to use.
        name : str = 'ApplyJitter'
            The name of the layer, which is used to index the layers dictionary.
        """
        super().__init__(name)
        self.kernel_size = int(kernel_size)
        self.sigma       = np.asarray(sigma, dtype=float)
        assert self.sigma.ndim == 0, ("sigma must be scalar array.")


    def generate_kernel(self : DetectorLayer) -> Array:
        """
        Generates the normalised guassian kernel.

        Returns
        -------
        kernel : Array
            The gaussian kernel.
        """
        # Generate distribution
        x = np.linspace(-10, 10, self.kernel_size)
        kernel = norm.pdf(x,          scale=self.sigma) * \
                 norm.pdf(x[:, None], scale=self.sigma)
        return kernel/np.sum(kernel)


    def __call__(self : DetectorLayer, image : Array) -> Array:
        """
        Convolves the input image with the generate gaussian kernel.

        Parameters
        ----------
        image : Array
            The image to convolve with the gussian kernel.

        Returns
        -------
        image : Array
            The image with the gaussian kernel convolution applied.
        """
        kernel = self.generate_kernel()
        return convolve(image, kernel, mode='same')
    

    def summary(self            : DetectorLayer, 
                angular_units   : str = 'radians', 
                cartesian_units : str = 'meters', 
                sigfigs         : int = 4) -> str:
        """
        Returns a summary of the class.

        Parameters
        ----------
        angular_units : str = 'radians'
            The angular units to use in the summary. Options are 'radians', 
            'degrees', 'arcseconds' and 'arcminutes'.
        cartesian_units : str = 'meters'
            The cartesian units to use in the summary. Options are 'meters',
            'millimeters' and 'microns'.
        sigfigs : int = 4
            The number of significant figures to use in the summary.

        Returns
        -------
        summary : str
            A summary of the class.
        """
        return (f"{self.name}: Applies a jitter of {self.jitter:.{sigfigs}} "
                "pixels.")


class ApplySaturation(DetectorLayer):
    """
    Applies a simple saturation model to the input image, by clipping any
    values above saturation, to saturation.

    Attributes
    ----------
    saturation : Array
        The value at which the saturation is applied.
    name : str
        The name of the layer, which is used to index the layers dictionary.
    """
    saturation : Array


    def __init__(self       : DetectorLayer,
                 saturation : Array,
                 name       : str = 'ApplySaturation') -> DetectorLayer:
        """
        Constructor for the ApplySaturation class.

        Parameters
        ----------
        saturation : Array
            The value at which the saturation is applied.
        name : str = 'ApplySaturation'
            The name of the layer, which is used to index the layers dictionary.
        """
        super().__init__(name)
        self.saturation = np.asarray(saturation, dtype=float)
        assert self.saturation.ndim == 0, ("saturation must be a scalar array.")


    def __call__(self : DetectorLayer, image : Array) -> Array:
        """
        Applies the satuation effect by reducing all values in the image above
        saturation, to the saturation value.

        Parameters
        ----------
        image : Array
            The image to apply the saturation to.

        Returns
        -------
        image : Array
            The image with the saturation applied.
        """
        return np.minimum(image, self.saturation)
    

    def summary(self            : DetectorLayer, 
                angular_units   : str = 'radians', 
                cartesian_units : str = 'meters', 
                sigfigs         : int = 4) -> str:
        """
        Returns a summary of the class.

        Parameters
        ----------
        angular_units : str = 'radians'
            The angular units to use in the summary. Options are 'radians', 
            'degrees', 'arcseconds' and 'arcminutes'.
        cartesian_units : str = 'meters'
            The cartesian units to use in the summary. Options are 'meters',
            'millimeters' and 'microns'.
        sigfigs : int = 4
            The number of significant figures to use in the summary.

        Returns
        -------
        summary : str
            A summary of the class.
        """
        return (f"{self.name}: Applies a saturation of "
                f"{self.saturation:.{sigfigs}} counts.")


class AddConstant(DetectorLayer):
    """
    Add a constant to the output image. This is typically used to model the
    mean value of the detector noise.

    Attributes
    ----------
    value : Array
        The value to add to the image.
    name : str
        The name of the layer, which is used to index the layers dictionary.
    """
    value : Array


    def __init__(self  : DetectorLayer,
                 value : Array,
                 name  : str = 'AddConstant') -> DetectorLayer:
        """
        Constructor for the AddConstant class.

        Parameters
        ----------
        value : Array
            The value to add to the image.
        name : str = 'AddConstant'
            The name of the layer, which is used to index the layers dictionary.
        """
        super().__init__(name)
        self.value = np.asarray(value, dtype=float)
        assert self.value.ndim == 0, ("value must be a scalar array.")


    def __call__(self : DetectorLayer, image : Array) -> Array:
        """
        Adds the value to the input image.

        Parameters
        ----------
        image : Array
            The image to add the value to.

        Returns
        -------
        image : Array
            The image with the value added.
        """
        return image + self.value
    

    def summary(self            : DetectorLayer, 
                angular_units   : str = 'radians', 
                cartesian_units : str = 'meters', 
                sigfigs         : int = 4) -> str:
        """
        Returns a summary of the class.

        Parameters
        ----------
        angular_units : str = 'radians'
            The angular units to use in the summary. Options are 'radians', 
            'degrees', 'arcseconds' and 'arcminutes'.
        cartesian_units : str = 'meters'
            The cartesian units to use in the summary. Options are 'meters',
            'millimeters' and 'microns'.
        sigfigs : int = 4
            The number of significant figures to use in the summary.

        Returns
        -------
        summary : str
            A summary of the class.
        """
        return (f"{self.name}: Adds a constant value of "
                f"{self.value:.{sigfigs}} counts.")


class IntegerDownsample(DetectorLayer):
    """
    Downsamples an input image by an integer number of pixels via a sum.
    The number of pixels in the input image must by integer divisible by the
    kernel_size.

    Attributes
    ----------
    kernel_size : int
        The size of the downsampling kernel.
    name : str
        The name of the layer, which is used to index the layers dictionary.
    """
    kernel_size : int


    def __init__(self        : DetectorLayer,
                 kernel_size : int,
                 name        : str = 'IntegerDownsample') -> DetectorLayer:
        """
        Constructor for the IntegerDownsample class.

        Parameters
        ----------
        kernel_size : int
            The size of the downsampling kernel.
        name : str = 'IntegerDownsample'
            The name of the layer, which is used to index the layers dictionary.
        """
        super().__init__(name)
        self.kernel_size = int(kernel_size)


    def downsample(self        : DetectorLayer,
                   array       : Array,
                   kernel_size : int) -> Array:
        """
        Downsamples the input array by kernel_size.

        Parameters
        ----------
        array : Array
            The input array to downsample.

        Returns
        -------
        kernel_size : int
            The size of the downsample kernel.
        """
        size_in = array.shape[0]
        size_out = size_in//kernel_size

        # Downsample first dimension
        array = array.reshape((size_in*size_out, kernel_size)).sum(1)
        array = array.reshape(size_in, size_out).T

        # Downsample second dimension
        array = array.reshape((size_out*size_out, kernel_size)).sum(1)
        array = array.reshape(size_out, size_out).T
        return array


    def __call__(self, image):
        """
        Downsamples the input image by the internally stored kernel_size.

        Parameters
        ----------
        image : Array
            The image to downsample.

        Returns
        -------
        image : Array
            The downsampled image.
        """
        return self.downsample(image, self.kernel_size)
    

    def summary(self            : DetectorLayer, 
                angular_units   : str = 'radians', 
                cartesian_units : str = 'meters', 
                sigfigs         : int = 4) -> str:
        """
        Returns a summary of the class.

        Parameters
        ----------
        angular_units : str = 'radians'
            The angular units to use in the summary. Options are 'radians', 
            'degrees', 'arcseconds' and 'arcminutes'.
        cartesian_units : str = 'meters'
            The cartesian units to use in the summary. Options are 'meters',
            'millimeters' and 'microns'.
        sigfigs : int = 4
            The number of significant figures to use in the summary.

        Returns
        -------
        summary : str
            A summary of the class.
        """
        return (f"{self.name}: Downsamples the image by {self.kernel_size} "
                f"pixels.")


class Rotate(DetectorLayer):
    """
    Applies a rotation to the image using interpolation methods.

    Parameters
    ----------
    angle : Array, radians
        The angle by which to rotate the image in the clockwise direction.
    fourier : bool
        Should the rotation be done using fourier methods or interpolation.
    padding : int
        The amount of padding to use if the fourier method is used.
    name : str
        The name of the layer, which is used to index the layers dictionary.
    """
    angle   : Array
    fourier : bool
    padding : int


    def __init__(self    : DetectorLayer,
                 angle   : Array,
                 fourier : bool = False,
                 padding : int  = None,
                 name    : str  = 'Rotate') -> DetectorLayer:
        """
        Constructor for the Rotate class.

        Parameters
        ----------
        angle: float, radians
            The angle by which to rotate the image in the clockwise direction.
        fourier : bool = False
            Should the fourier rotation method be used (True), or regular
            interpolation method be used (False).
        padding : int = None
            The amount of fourier padding to use. Only applies if fourier is
            True.
        name : str = 'Rotate'
            The name of the layer, which is used to index the layers dictionary.
        """
        super().__init__(name)
        self.angle   = np.asarray(angle, dtype=float)
        self.fourier = bool(fourier)
        self.padding = padding if padding is None else int(padding)
        assert self.angle.ndim == 0, ("angle must be scalar array.")


    def __call__(self : DetectorLayer, image : Array) -> Array:
        """
        Applies the rotation to an image.

        Parameters
        ----------
        image : Array
            The image to rotate.

        Returns
        -------
        image : Array
            The rotated image.
        """
        if self.fourier:
            return fourier_rotate(image, self.angle, self.padding)
        else:
            return rotate(image, self.angle)
        

    def summary(self            : OpticalLayer, 
                angular_units   : str = 'radians', 
                cartesian_units : str = 'meters', 
                sigfigs         : int = 4) -> str:
        """
        Returns a summary of the class.

        Parameters
        ----------
        angular_units : str = 'radians'
            The angular units to use in the summary. Options are 'radians', 
            'degrees', 'arcseconds' and 'arcminutes'.
        cartesian_units : str = 'meters'
            The cartesian units to use in the summary. Options are 'meters',
            'millimeters' and 'microns'.
        sigfigs : int = 4
            The number of significant figures to use in the summary.

        Returns
        -------
        summary : str
            A summary of the class.
        """
        angle = convert_angular(self.angle, 'radians', angular_units)
        
        if self.fourier:
            method = f"a Fourier method with padding of {self.padding}"
        else:
            method = "an Interpolation method of order 1"

        return (f"{self.name}: Applies a {angle:.{sigfigs}} {angular_units} "
                f"rotation to the image using {method}.")


from __future__ import annotations
import jax.numpy as np
import pytest
import dLux
from jax import config
config.update("jax_debug_nans", True)

Array = np.ndarray


class TestCreateWavefront(object):
    """
    Tests the CreateWavefront class.
    """


    def test_constructor(self, create_create_wavefront: callable) -> None:
        """
        Tests the constructor.
        """
        # Test wrong dims
        with pytest.raises(AssertionError):
            create_create_wavefront(diameter=np.array([]))

        # Test wrong string input
        with pytest.raises(AssertionError):
            create_create_wavefront(wavefront_type='cartesian')

        # Test functioning
        create_create_wavefront()


    def test_call(self, create_create_wavefront: callable) -> None:
        """
        Tests the __call__ method.
        """
        params = {'wavelength': np.array(1e-6), 'offset': np.zeros(2)}
        create_create_wavefront()(None, params)


class TestTiltWavefront(object):
    """
    Tests the TiltWavefront class.
    """


    def test_constructor(self, create_tilt_wavefront: callable) -> None:
        """
        Tests the constructor.
        """
        # Test wrong dims
        with pytest.raises(AssertionError):
            create_tilt_wavefront(tilt_angles=np.ones(1))

        # Test wrong dims
        with pytest.raises(AssertionError):
            create_tilt_wavefront(tilt_angles=np.array([]))

        # Test functioning
        create_tilt_wavefront()


    def test_call(self, 
            create_tilt_wavefront: callable,
            create_wavefront: callable) -> None:
        """
        Tests the __call__ method.
        """
        wf = create_wavefront() 
        create_tilt_wavefront()(wf)


class TestNormaliseWavefront(object):
    """
    Tests the NormaliseWavefront class.
    """


    def test_constructor(self, create_normalise_wavefront: callable) -> None:
        """
        Tests the constructor.
        """
        # Test functioning
        create_normalise_wavefront()


    def test_call(self, 
            create_normalise_wavefront: callable,
            create_wavefront: callable) -> None:
        """
        Tests the __call__ method.
        """
        wf = create_wavefront() 
        wf = create_normalise_wavefront()(wf)
        assert wf.psf.sum() == 1.


class TestApplyBasisOPD(object):
    """
    Tests the ApplyBasisOPD class.
    """


    def test_constructor(self, create_apply_basis_opd: callable) -> None:
        """
        Tests the constructor.
        """
        # Test wrong dims
        with pytest.raises(AssertionError):
            create_apply_basis_opd(basis=np.ones((16, 16)))

        # Test wrong dims
        with pytest.raises(AssertionError):
            create_apply_basis_opd(basis=np.ones((1, 1, 16, 16)))

        # Test wrong dims
        with pytest.raises(AssertionError):
            create_apply_basis_opd(coefficients=np.array([]))

        # Test wrong dims
        with pytest.raises(AssertionError):
            create_apply_basis_opd(coefficients=np.zeros((1, 1)))

        # Test wrong dims
        with pytest.raises(AssertionError):
            create_apply_basis_opd(basis=np.ones((2, 15, 15)),
                                   coefficients=np.zeros((3)))

        # Test functioning
        create_apply_basis_opd()


    def test_call(self, 
            create_apply_basis_opd: callable, 
            create_wavefront: callable) -> None:
        """
        Tests the __call__ method.
        """
        wf = create_wavefront() 
        create_apply_basis_opd()(wf)


class TestAddPhase(object):
    """
    Tests the AddPhase class.
    """


    def test_constructor(self, create_add_phase: callable) -> None:
        """
        Tests the constructor.
        """
        # Test wrong dims
        with pytest.raises(AssertionError):
            create_add_phase(phase=np.ones(1))

        # Test functioning
        create_add_phase()


    def test_call(self, 
            create_add_phase: callable, 
            create_wavefront: callable) -> None:
        """
        Tests the __call__ method.
        """
        wf = create_wavefront() 
        npix = wf.npixels

        # Test 0d
        create_add_phase(phase=np.array(1.))(wf)

        # Test 2d
        create_add_phase(phase=np.ones((npix, npix)))(wf)

        # Test 3d
        create_add_phase(phase=np.ones((1, npix, npix)))(wf)


class TestAddOPD(object):
    """
    Tests the AddOPD class.
    """


    def test_constructor(self, create_add_opd: callable) -> None:
        """
        Tests the constructor.
        """
        # Test wrong dims
        with pytest.raises(AssertionError):
            create_add_opd(opd=np.ones(1))

        # Test functioning
        create_add_opd()


    def test_call(self, 
            create_add_opd: callable, 
            create_wavefront: callable) -> None:
        """
        Tests the __call__ method.
        """
        wf = create_wavefront() 
        npix = wf.npixels

        # Test 0d
        create_add_opd(opd=np.array(1.))(wf)

        # Test 2d
        create_add_opd(opd=np.ones((npix, npix)))(wf)

        # Test 3d
        create_add_opd(opd=np.ones((1, npix, npix)))(wf)


class TestTransmissiveOptic(object):
    """
    Tests the TransmissiveOptic class.
    """


    def test_constructor(self, create_transmissive_optic: callable) -> None:
        """
        Tests the constructor.
        """
        # Test wrong dims
        with pytest.raises(AssertionError):
            create_transmissive_optic(trans=np.ones(1))

        # Test functioning
        create_transmissive_optic()


    def test_call(self, 
            create_transmissive_optic: callable, 
            create_wavefront: callable) -> None:
        """
        Tests the __call__ method.
        """
        wf = create_wavefront() 
        npix = wf.npixels

        # Test 0d
        create_transmissive_optic(trans=np.array(1.))(wf)

        # Test 2d
        create_transmissive_optic(trans=np.ones((npix, npix)))(wf)

        # Test 3d
        create_transmissive_optic(trans=np.ones((1, npix, npix)))(wf)


class TestApplyBasisCLIMB(object):
    """
    Tests the ApplyBasisCLIMB class.
    """


    def test_constructor(self, create_basis_climb: callable) -> None:
        """
        Tests the constructor.
        """
        # Test wrong dims
        with pytest.raises(AssertionError):
            create_basis_climb(basis=np.ones((16, 16)))

        # Test wrong dims
        with pytest.raises(AssertionError):
            create_basis_climb(basis=np.ones((1, 1, 16, 16)))

        # Test wrong dims
        with pytest.raises(AssertionError):
            create_basis_climb(ideal_wavelength=np.ones(1))

        # Test wrong dims
        with pytest.raises(AssertionError):
            create_basis_climb(coefficients=np.array([]))

        # Test wrong dims
        with pytest.raises(AssertionError):
            create_basis_climb(coefficients=np.zeros((1, 1)))

        # Test wrong dims
        with pytest.raises(AssertionError):
            create_basis_climb(basis=np.ones((2, 15, 15)),
                                   coefficients=np.zeros((3)))

        # Test functioning
        create_basis_climb()


    def test_call(self, 
            create_basis_climb: callable, 
            create_wavefront: callable) -> None:
        """
        Tests the __call__ method.
        """
        wf = create_wavefront(amplitude = np.ones((1, 256, 256)),
                            phase = np.zeros((1, 256, 256))) 
        npix = wf.npixels
        basis = np.ones((3, 3*npix, 3*npix))
        create_basis_climb(basis=basis)(wf)


class TestRotate(object):
    """
    Tests the Rotate class.
    """


    def test_constructor(self, create_rotate: callable) -> None:
        """
        Tests the constructor.
        """
        # Test wrong dims
        with pytest.raises(AssertionError):
            create_rotate(angle=np.ones(1))

        # Test functioning
        create_rotate()


    def test_call(self, 
            create_rotate: callable, 
            create_wavefront: callable) -> None:
        """
        Tests the __call__ method.
        """
        # Test regular rotation
        wf = create_wavefront()
        create_rotate()(wf)

        # Test real imaginary rotation
        wf = create_wavefront()
        create_rotate(real_imaginary=True)(wf)

        # Test fourier
        with pytest.raises(NotImplementedError):
            wf = create_wavefront()
            create_rotate(fourier=True)(wf)



from __future__ import annotations
from zodiax import Base
import jax.numpy as np
from jax import lax
import jax.tree_util as jtu
import dLux
from dLux.utils.math import factorial
from dLux.utils.coordinates import cartesian_to_polar, get_pixel_positions


__all__ = ['Zernike', 'ZernikeBasis', 'AberrationFactory']


Array = np.ndarray


zernike_names = {
    # 0th Radial
    1: 'Piston',

    # 1st Radial
    2: 'Tilt X',
    3: 'Tilt Y',
    
    # Second Radial
    4: 'Defocus',
    5: 'Astig X',
    6: 'Astig Y',
    
    # Third Radial
    7: 'Coma X',
    8: 'Coma Y',
    9: 'Trefoil X',
    10: 'Trefoil Y',

    # Fourth Radial
    11: 'Spherical',
    12: '2nd Astig X',
    13: '2nd Astig Y',
    14: 'Quadrafoil X',
    15: 'Quadrafoil Y',
    
    # Fifth Radial
    16: '2nd Coma X',
    17: '2nd Coma Y',
    18: '2nd Trefoil X',
    19: '2nd Trefoil Y',
    20: 'Pentafoil X',
    21: 'Pentafoil Y',
    
    # Sixth Radial
    22: '2nd Spherical',
    23: '3rd Coma X',
    24: '3rd Coma Y',
    25: '3rd Astig X',
    26: '3rd Astig Y',
    27: 'Hexafoil X',
    28: 'Hexafoil Y',
    
    # Seventh Radial
    29: '4th Coma X',
    30: '4th Coma Y',
    31: '4th Astig X',
    32: '4th Astig Y',
    33: '3rd Trefoil X',
    34: '3rd Trefoil Y',
    35: 'Heptafoil X',
    36: 'Heptafoil Y',
}


class Zernike(Base):
    """
    A class to generate Zernike polynomials dynamically.

    The 'jth' zernike polynomial is defined [here](https://oeis.org/A176988).
    The basic translation between the noll index and the pair of numbers is
    shown below:

    1 -> (0, 0)

    2, 3 -> (1, -1), (1, 1)

    4, 5, 6 -> (2, -2), (2, 0), (2, 2)

    7, 8, 9, 10 -> (3, -3), (3, -1), (3, 1), (3, 3)

    
    Attributes
    ----------
    j : int
        The Zernike (noll) index.
    n : int
        The radial order of the Zernike polynomial.
    m : int
        The azimuthal order of the Zernike polynomial.
    name : str
        The name of the Zernike polynomial.
    _k : Array
        The array of powers using the radial calculation. This is a 
        pre-calculated parameter and should not be changed.
    _c : Array
        The array of normalisaton coefficients used in the radial calculation.
        This is a pre-calculated parameter and should not be changed.
    """
    j    : int
    n    : int
    m    : int
    name : str
    _k    : Array
    _c    : Array

    def __init__(self : Zernike, j : int):
        """
        Construct for the Zernike class.

        Parameters
        ----------
        j : int
            The Zernike (noll) index.
        """
        if int(j) < 1:
            raise ValueError('The Zernike index must be greater than 0.')
        self.j = int(j)
        self.n, self.m = self._noll_index(self.j)
        self.name = zernike_names[int(self.j)] if self.j >= 1 and self.j <= 36 \
                    else f'Zernike {int(self.j)}'

        # Calcualte values
        self._k = np.arange(((self.n - self.m) // 2) + 1, dtype=float)
        sign = lax.pow(-1., self._k)
        _fact_1 = factorial(np.abs(self.n - self._k))
        _fact_2 = factorial(self._k)
        _fact_3 = factorial(((self.n + self.m) // 2) - self._k)
        _fact_4 = factorial(((self.n - self.m) // 2) - self._k)
        self._c = sign * _fact_1 / _fact_2 / _fact_3 / _fact_4 


    def _noll_index(self : Zernike, j : int) -> tuple[int]:
        """
        Calculate the radial and azimuthal orders of the Zernike polynomial.

        Parameters
        ----------
        j : int
            The Zernike (noll) index.

        Returns
        -------
        n, m : tuple[int]
            The radial and azimuthal orders of the Zernike polynomial.
        """
        n = (np.ceil(-1 / 2 + np.sqrt(1 + 8 * j) / 2) - 1).astype(int)
        smallest_j_in_row = n * (n + 1) / 2 + 1 
        number_of_shifts = (j - smallest_j_in_row + ~(n & 1) + 2) // 2
        sign_of_shift = -(j & 1) + ~(j & 1) + 2
        base_case = (n & 1)
        m = (sign_of_shift * (base_case + number_of_shifts * 2)).astype(int)
        return int(n), int(m)
    

    def _calculate_radial(self : Zernike, rho : Array) -> Array:
        """
        Calculates the radial component of the Zernike polynomial.

        Parameters
        ----------
        rho : Array
            The radial coordinate of the Zernike polynomial.
        
        Returns
        -------
        radial : Array
            The radial component of the Zernike polynomial.
        """
        rads = lax.pow(rho[:, :, None], 
            (np.abs(self.n) - 2 * self._k)[None, None, :])
        return (self._c * rads).sum(axis = 2)


    def _calculate_azimuthal(self : Zernike, theta : Array) -> Array:
        """
        Calculates the azimuthal component of the Zernike polynomial.

        Parameters
        ----------
        theta : Array
            The azimuthal coordinate of the Zernike polynomial.

        Returns
        -------
        azimuthal : Array
            The azimuthal component of the Zernike polynomial.
        """
        norm_coeff = np.sqrt(self.n + 1)
        if self.m != 0:
            norm_coeff *= 1 + (np.sqrt(2) - 1)
        
        if self.m >= 0:
            return norm_coeff * np.cos(np.abs(self.m) * theta)
        else:
            return norm_coeff * np.sin(np.abs(self.m) * theta)
        

    def calculate_zernike(self : Zernike, coordinates : Array) -> Array:
        """
        Calculates the Zernike polynomial.

        Note: The zernike polynomial is defined on the coordinates up to a
        radial value of 1.

        Parameters
        ----------
        coordinates : Array
            The cartesian coordinates to calcualte the Zernike polynomial upon.
        
        Returns
        -------
        zernike : Array
            The Zernike polynomial.
        """
        polar_coordinates = cartesian_to_polar(coordinates)
        rho = polar_coordinates[0]
        theta = polar_coordinates[1]
        aperture = rho <= 1.
        return aperture * self._calculate_radial(rho) * \
            self._calculate_azimuthal(theta)
    

    def calculate_polike(self        : Zernike, 
                         coordinates : Array, 
                         nsides      : int) -> Array:
        """
        Calculates the Zernike polynomial on an nsided aperture.

        Note: The zernike polynomial is defined on the coordinates up to a
        radial value of 1.

        Parameters
        ----------
        coordinates : Array
            The cartesian coordinates to calcualte the Zernike polynomial upon.
        nsides : int
            The number of sides of the aperture.

        Returns
        -------
        polike : Array
            The Zernike polynomial on an nsided aperture.
        """
        if nsides < 3:
            raise ValueError(f'nsides must be >= 3, not {nsides}.')
        theta = cartesian_to_polar(coordinates)[1]
        alpha = np.pi / nsides
        phi = theta + alpha  
        wedge = np.floor((phi + alpha) / (2. * alpha))
        u_alpha = phi - wedge * (2 * alpha)
        r_alpha = np.cos(alpha) / np.cos(u_alpha)
        return 1 / r_alpha * self.calculate_zernike(coordinates / r_alpha)


    def calculate(self        : Zernike, 
                  coordinates : Array, 
                  nsides      : int = 0) -> Array:
        """
        Calculates the Zernike polynomial.

        Note: The zernike polynomial is defined on the coordinates up to a
        radial value of 1.

        Parameters
        ----------
        coordinates : Array
            The cartesian coordinates to calcualte the Zernike polynomial upon.
        nsides : int
            The number of sides of the aperture. If 0, the Zernike polynomial
            is calculated on a circular aperture.

        Returns
        -------
        zernike : Array
            The Zernike polynomial.
        """
        if nsides == 0:
            return self.calculate_zernike(coordinates)
        else:
            return self.calculate_polike(coordinates, nsides)

    
class ZernikeBasis(Base):
    """
    A class to calculate a set of Zernike polynomials on a dynamic set of
    coordiantes.

    The 'jth' zernike polynomial is defined [here](https://oeis.org/A176988).
    The basic translation between the noll index and the pair of numbers is
    shown below:
    
    1 -> (0, 0)
    
    2, 3 -> (1, -1), (1, 1)
    
    4, 5, 6 -> (2, -2), (2, 0), (2, 2)
    
    7, 8, 9, 10 -> (3, -3), (3, -1), (3, 1), (3, 3)

    Attributes
    ----------
    zernikes : list[Zernike]
        The list of Zernike polynomial classes to calculate.
    """
    zernikes : list[Zernike]


    def __init__(self : ZernikeBasis, js : list[int]):
        """
        Constructor for the DynamicZernike class.

        Parameters
        ----------
        js : list[int]
            The list of Zernike (noll) indices to calculate.
        """
        self.zernikes = [Zernike(j) for j in js]
    

    def calculate_basis(self        : ZernikeBasis, 
                        coordinates : Array, 
                        nsides      : int = 0) -> Array:
        """
        Calculates the full Zernike polynomial basis.

        Note: The zernike polynomial is defined on the coordinates up to a
        radial value of 1.

        Parameters
        ----------
        coordinates : Arraya
            The cartesian coordinates to calcualte the Zernike basis upon.
        
        Returns
        -------
        zernike_basis : Array
            The Zernike polynomial basis.
        """
        leaf_fn = lambda leaf: isinstance(leaf, Zernike)
        calculate_fn = lambda z: z.calculate(coordinates, nsides)
        return np.array(jtu.tree_map(calculate_fn, self.zernikes, 
                                     is_leaf=leaf_fn))



class AberrationFactory():
    """
    This class is not actually ever instatiated, but is rather a class used to 
    give a simple constructor interface that is used to construct the most
    commonly used aberrations. It is able to construct hard-edged circular or 
    regular poygonal aberrations. 

    Lets look at an example of how to construct a simple circular aberration 
    class. Let calcualte this for a 512x512 array with the aperture spanning
    the full array.

    ```python
    from dLux import AberrationFactory
    import jax.numpy as np
    import jax.random as jr
    
    # Construct Zernikes
    zernikes = np.arange(4, 11)
    coefficients = jr.normal(jr.PRNGKey(0), (zernikes.shape[0],))

    # Construct aberrations
    aberrations = AberrationFactory(512, zernikes=zernikes, 
                                    coefficients=coefficients)
    ```
    
    The resulting aperture class has two parameters, `.basis` and 
    `.coefficients`. We can then examine the opd like so:

    ```python
    import matplotlib.pyplot as plt
    plt.figure(figsize=(5, 4))
    plt.imshow(aberrations.get_opd())
    plt.colorbar()
    plt.show()
    ```

    We can also easily change this to a hexagonal aperture:

    ```python
    # Construct aberrations
    aberrations = AberrationFactory(512, nsides=6, zernikes=zernikes, 
                                    coefficients=coefficients)
    
    # Examine
    plt.figure(figsize=(5, 4))
    plt.imshow(aberrations.get_opd())
    plt.colorbar()
    plt.show()
    ```
    """
    def __new__(cls              : AberrationFactory, 
                npixels          : int, 
                nsides           : int   = 0,
                rotation         : float = 0., 

                # Sizing
                aperutre_ratio   : float = 1.0,

                # Aberrations
                zernikes         : Array = None, 
                coefficients     : Array = None, 

                # name
                name             : str = None):
        """
        Constructs a basic single static aberration class.

        TODO: Add link to the zenike noll indicies

        Parameters
        ----------
        npixels : int
            Number of pixels used to represent the aperture.
        nsides : int = 0
            Number of sides of the aperture. A zero input results in a circular
            aperture. All other other values of three and above are supported.
        rotation : float, radians = 0
            The global rotation of the aperture in radians.
        aperutre_ratio : float = 1.
            The ratio of the aperture size to the array size. A value of 1. 
            results in an aperture that fully spans the array, a value of 0.5 
            retuls in an aperure that is half the size of the array, which is 
            equivilent to a padding factor of 2.
        zernikes : Array = None
            The zernike noll indices to be used for the aberrations. Please 
            refer to (this)[Add this link] docstring to see which indicides 
            correspond to which aberrations. Typical values are range(4, 11).
        coefficients : Array = None
            The zernike cofficients to be applied to the aberrations. Defaults 
            to an array of zeros.
        name : str = None
            The name of the aperture used to index the layers dictionary. If 
            not supplied, the aperture will be named based on the number of
            sides. However this is only supported up to 8 sides, and a name
            must be supplied for apertures with more than 8 sides.
        
        Returns
        -------
        aperture : ApplyBasisOPD
            Returns an appropriately constructed ApplyBasisOPD.
        """
        # Check vaid inputs
        if nsides < 3 and nsides != 0:
            raise ValueError("nsides must be either 0 or >=3")
        
        # Auto-name
        if name is None:
            if nsides > 8:
                raise ValueError("Warning: Auto-naming not supported for " + \
                "nsides > 8. Please provide a name.")
            sides = ["Circular", "Triangular", "Square", "Pentagonal", 
                "Hexagonal", "Heptagonal", "Octagonal"]
            name = sides[np.maximum(nsides-2, 0)] + "Aperture"


        # Construct coordinates
        coords = get_pixel_positions((npixels, npixels), (1/npixels, 1/npixels))

        # Circular Primary
        if nsides == 0:
            ap = dLux.apertures.CircularAperture
            dyn_aperture = ap(aperutre_ratio/2, softening=0)
            coordinates = dyn_aperture._normalised_coordinates(coords)

        # Polygonal Primary
        else: 
            ap = dLux.apertures.RegularPolygonalAperture
            dyn_aperture = ap(nsides, aperutre_ratio/2, softening=0, 
                          rotation=rotation)
            coordinates = dyn_aperture._normalised_coordinates(coords, nsides)

        # Construct Aberrations
        basis = ZernikeBasis(zernikes).calculate_basis(coordinates)
        return dLux.optics.ApplyBasisOPD(basis, coefficients, name=name)


    def __init__(self              : AberrationFactory, 
                npixels          : int, 
                nsides           : int   = 0,
                rotation         : float = 0., 

                # Sizing
                aperutre_ratio   : float = 1.0,

                # Aberrations
                zernikes         : Array = None, 
                coefficients     : Array = None, 

                # name
                name             : str = None):
        """
        Constructs a basic single static aberration class.

        Parameters
        ----------
        npixels : int
            Number of pixels used to represent the aperture.
        nsides : int = 0
            Number of sides of the aperture. A zero input results in a circular
            aperture. All other other values of three and above are supported.
        rotation : float, radians = 0
            The global rotation of the aperture in radians.
        aperutre_ratio : float = 1.
            The ratio of the aperture size to the array size. A value of 1. 
            results in an aperture that fully spans the array, a value of 0.5 
            retuls in an aperure that is half the size of the array, which is 
            equivilent to a padding factor of 2.
        zernikes : Array = None
            The zernike noll indices to be used for the aberrations. Please 
            refer to (this)[Add this link] docstring to see which indicides 
            correspond to which aberrations. Typical values are range(4, 11).
        coefficients : Array = None
            The zernike cofficients to be applied to the aberrations. Defaults 
            to an array of zeros.
        name : str = None
            The name of the aperture used to index the layers dictionary. If 
            not supplied, the aperture will be named based on the number of
            sides. However this is only supported up to 8 sides, and a name
            must be supplied for apertures with more than 8 sides.
        
        Returns
        -------
        aperture : ApplyBasisOPD
            Returns an appropriately constructed ApplyBasisOPD.
        """



import jax.numpy as np
from jax import hessian, grad
from jax.scipy.stats import poisson
from typing import Callable

__all__ = ["poisson_likelihood", "poisson_log_likelihood",
           "chi2_likelihood", "chi2_log_likelihood",
           "calculate_covariance", "calculate_entropy"]


Array = np.ndarray


def poisson_likelihood(X         : Array,
                       data      : Array,
                       model     : object,
                       update_fn : Callable,
                       model_fn  : str,
                       *args,
                       **kwargs) -> Array:
    """
    Calculates the poissonion likelihood after updating the model using the
    update_fn. The extra *args and **kwargs are passed to the model function.

    Parameters
    ----------
    X : Array
        The array of values to update the model with.
    data : Array
        The data to calulcate the prior with respect to.
    model : object
        The model used to model the data.
    update_fn : Callable
        The function used to update the model before modelling. This should
        take in the X array as the first argument and the model as the second.
    model_fn : str
        The name of the model function to use.

    Returns
    -------
    likelhood : Array
        The poissonian likelihood of the updated model with respect to the data.
    """
    model = update_fn(X, model)
    psf = getattr(model, model_fn)(*args, **kwargs)
    return np.sum(poisson.pmf(data, psf))


def poisson_log_likelihood(X         : Array,
                           data      : Array,
                           model     : object,
                           update_fn : Callable,
                           model_fn  : str,
                           *args,
                           **kwargs) -> Array:
    """
    Calculates the poissonion log likelihood after updating the model using the
    update_fn. The extra *args and **kwargs are passed to the model function.

    Parameters
    ----------
    X : Array
        The array of values to update the model with.
    data : Array
        The data to calulcate the prior with respect to.
    model : object
        The model used to model the data.
    update_fn : Callable
        The function used to update the model before modelling. This should
        take in the X array as the first argument and the model as the second.
    model_fn : str
        The name of the model function to use.

    Returns
    -------
    likelhood : Array
        The poissonian log likelihood of the updated model with respect to the
        data.
    """
    model = update_fn(X, model)
    psf = getattr(model, model_fn)(*args, **kwargs)
    return np.sum(poisson.logpmf(data, psf))


def chi2_likelihood(X         : Array,
                    data      : Array,
                    model     : object,
                    update_fn : Callable,
                    model_fn  : str,
                    *args,
                    **kwargs) -> Array:
    """
    Calculates the chi2 likelihood after updating the model using the
    update_fn. The extra *args and **kwargs are passed to the model function.

    Parameters
    ----------
    X : Array
        The array of values to update the model with.
    data : Array
        The data to calulcate the prior with respect to.
    model : object
        The model used to model the data.
    update_fn : Callable
        The function used to update the model before modelling. This should
        take in the X array as the first argument and the model as the second.
    model_fn : str
        The name of the model function to use.

    Returns
    -------
    likelhood : Array
        The chi2 likelihood of the updated model with respect to the data.
    """
    model = update_fn(X, model)
    psf = getattr(model, model_fn)(*args, **kwargs)
    return np.sum(np.square((psf-data)/noise))


def chi2_log_likelihood(X         : Array,
                        data      : Array,
                        model     : object,
                        update_fn : Callable,
                        model_fn  : str,
                        *args,
                        **kwargs) -> Array:
    """
    Calculates the chi2 log likelihood after updating the model using the
    update_fn. The extra *args and **kwargs are passed to the model function.

    Parameters
    ----------
    X : Array
        The array of values to update the model with.
    data : Array
        The data to calulcate the prior with respect to.
    model : object
        The model used to model the data.
    update_fn : Callable
        The function used to update the model before modelling. This should
        take in the X array as the first argument and the model as the second.
    model_fn : str
        The name of the model function to use.

    Returns
    -------
    likelhood : Array
        The chi2 log likelihood of the updated model with respect to the data.
    """
    model = update_fn(X, model)
    psf = getattr(model, model_fn)(*args, **kwargs)
    return np.log(np.sum(np.square((psf-data)/noise)))


def calculate_covariance(likelihood_fn : Callable,
                         *args,
                         **kwargs) -> Array:
    """
    Calcuates the covariance matrix under the Laplace approximation for the
    given likelihood function. The *args and **kwargs are passed to the
    likelihood function.

    Parameters
    ----------
    likelihood_fn : Callable
        The likelihood function to calculate the covaraince matrix with
        respect to.

    Returns
    -------
    covaraince : Array
        The corresponding covariance matrix.
    """
    matrix = hessian(likelihood_fn)(*args, **kwargs)
    return -np.linalg.inv(matrix)


def calculate_entropy(likelihood_fn : Callable,
                      *args,
                      **kwargs):
    """
    Calcuates the entropy of the covaraince matrix under the Laplace
    approximation for the given likelihood function. The *args and **kwargs
    are passed to through the calculate_covaraince function through to the
    likelihood function.

    Parameters
    ----------
    likelihood_fn : Callable
        The likelihood function to calculate the entropy with respect to.

    Returns
    -------
    entropy : Array
        The entropy of the covariance matrix for the given likelihood function.
    """
    cov = calculate_covariance(likelihood_fn, *args, **kwargs)
    sign, logdet = np.linalg.slogdet(cov)
    return 0.5 * (np.log(2 * np.pi * np.e) + (sign * logdet))


import jax.numpy as np
import pytest
import dLux
from jax import config
config.update("jax_debug_nans", True)

Array = np.ndarray

class TestWavefront(object):
    """
    Test the Wavefront class.
    """


    def test_constructor(self, create_wavefront: callable) -> None:
        """
        Tests the constructor.
        """
        # Test constructor
        create_wavefront()

        # Test empty array
        with pytest.raises(AssertionError):
            create_wavefront(wavelength=[])

        # Test 1d array
        with pytest.raises(AssertionError):
            create_wavefront(wavelength=[1e3, 1e4])

        # Test empty array
        with pytest.raises(AssertionError):
            create_wavefront(pixel_scale=[])

        # Test 1d array
        with pytest.raises(AssertionError):
            create_wavefront(pixel_scale=[1e3, 1e4])

        # Test non 3d amplitude array
        with pytest.raises(AssertionError):
            create_wavefront(amplitude=np.ones((3, 3)))

        # Test non 3d amplitude array
        with pytest.raises(AssertionError):
            create_wavefront(amplitude=np.ones((3, 3, 3, 3)))

        # Test non 3d phase array
        with pytest.raises(AssertionError):
            create_wavefront(phase=np.ones((3, 3)))

        # Test non 3d phase array
        with pytest.raises(AssertionError):
            create_wavefront(phase=np.ones((3, 3, 3, 3)))

        # Test different amplitude/phase array shapes
        with pytest.raises(AssertionError):
            create_wavefront(amplitude=np.ones((4, 4, 4)),
                                   phase=np.ones((3, 3, 3)))

        # Test non-planetype plane_type
        with pytest.raises(AssertionError):
            create_wavefront(plane_type=[1])


    def test_npixels(self, create_wavefront: callable) -> None:
        """
        Tests the npixels property.
        """
        wf = create_wavefront()
        assert wf.npixels == wf.amplitude.shape[-1]


    def test_nfields(self, create_wavefront: callable) -> None:
        """
        Tests the nfields property.
        """
        wf = create_wavefront()
        assert wf.nfields == wf.amplitude.shape[0]


    def test_diameter(self, create_wavefront: callable) -> None:
        """
        Tests the diameter property.
        """
        wf = create_wavefront()
        assert wf.diameter == wf.npixels * wf.pixel_scale


    def test_real(self, create_wavefront: callable) -> None:
        """
        Tests the real property.
        """
        wf = create_wavefront()
        assert (wf.real == wf.amplitude * np.cos(wf.phase)).all()


    def test_imaginary(self, create_wavefront: callable) -> None:
        """
        Tests the imaginary property.
        """
        wf = create_wavefront()
        assert (wf.imaginary == wf.amplitude * np.sin(wf.phase)).all()


    def test_phasor(self, create_wavefront: callable) -> None:
        """
        Tests the phasor property.
        """
        wf = create_wavefront()
        assert (wf.phasor == wf.amplitude * np.exp(1j*wf.phase)).all()


    def test_psf(self, create_wavefront: callable) -> None:
        """
        Tests the psf property.
        """
        wf = create_wavefront()
        assert (wf.psf == wf.amplitude**2).all()


    def test_pixel_coordinates(self, create_wavefront: callable) -> None:
        """
        Tests the pixel_coordinates property.
        """
        wf = create_wavefront()
        assert (wf.pixel_coordinates == \
        dLux.utils.coordinates.get_pixel_positions((wf.npixels, wf.npixels,),
                                        (wf.pixel_scale, wf.pixel_scale))).all()


    def test_set_amplitude(self, create_wavefront: callable) -> None:
        """
        Tests the set_amplitude method.
        """
        wf = create_wavefront()

        # Test string inputs
        with pytest.raises(AssertionError):
            wf.set_amplitude("some string")

        # Test list inputs
        with pytest.raises(AssertionError):
            wf.set_amplitude([])

        # Test wrong shapes
        with pytest.raises(AssertionError):
            wf.set_amplitude(np.ones((16, 16)))

        # Test wrong shapes
        with pytest.raises(AssertionError):
            wf.set_amplitude(np.ones((1, 1, 16, 16)))

        # Test correct behaviour
        new_ampl = 0.5*np.ones((1, 16, 16))
        assert (wf.set_amplitude(new_ampl).amplitude == new_ampl).all()


    def test_set_phase(self, create_wavefront: callable) -> None:
        """
        Tests the set_phase method.
        """
        wf = create_wavefront()

        # Test string inputs
        with pytest.raises(AssertionError):
            wf.set_phase("some string")

        # Test list inputs
        with pytest.raises(AssertionError):
            wf.set_phase([])

        # Test wrong shapes
        with pytest.raises(AssertionError):
            wf.set_phase(np.ones((16, 16)))

        # Test wrong shapes
        with pytest.raises(AssertionError):
            wf.set_phase(np.ones((1, 1, 16, 16)))

        # Test correct behaviour
        new_phase = 0.5*np.ones((1, 16, 16))
        assert (wf.set_phase(new_phase).phase == new_phase).all()


    def test_set_pixel_scale(self, create_wavefront: callable) -> None:
        """
        Tests the set_pixel_scale method.
        """
        wf = create_wavefront()

        # Test string inputs
        with pytest.raises(AssertionError):
            wf.set_pixel_scale("some string")

        # Test list inputs
        with pytest.raises(AssertionError):
            wf.set_pixel_scale([])

        # Test wrong shapes
        with pytest.raises(AssertionError):
            wf.set_pixel_scale(np.array([]))

        # Test correct behaviour
        new_pixscale = np.array(1.5)
        assert wf.set_pixel_scale(new_pixscale).pixel_scale == new_pixscale


    def test_set_plane_type(self, create_wavefront: callable) -> None:
        """
        Tests the set_plane_type method.
        """
        wf = create_wavefront()

        # Test string inputs
        with pytest.raises(AssertionError):
            wf.set_plane_type("some string")

        # Test correct behaviour
        new_plane_type = dLux.PlaneType.Focal
        assert wf.set_plane_type(new_plane_type).plane_type == new_plane_type


    def test_set_phasor(self, create_wavefront: callable) -> None:
        """
        Tests the set_phasor method.
        """
        wf = create_wavefront()

        # Test string inputs
        with pytest.raises(AssertionError):
            wf.set_phasor("some string", "some_string")

        # Test list inputs
        with pytest.raises(AssertionError):
            wf.set_phasor([], [])

        # Test wrong shapes
        with pytest.raises(AssertionError):
            wf.set_phasor(np.ones((16, 16)), np.ones((16, 16)))

        # Test wrong shapes
        with pytest.raises(AssertionError):
            wf.set_phasor(np.ones((1, 1, 16, 16)), np.ones((1, 1, 16, 16)))

        # Test wrong shapes
        with pytest.raises(AssertionError):
            wf.set_phasor(np.ones((1, 16, 16)), np.ones((1, 15, 15)))

        # Test correct behaviour
        new_ampl = 0.5*np.ones((1, 16, 16))
        new_phase = 0.5*np.ones((1, 16, 16))
        assert (wf.set_phasor(new_ampl, new_phase).amplitude == new_ampl).all()
        assert (wf.set_phasor(new_ampl, new_phase).phase == new_phase).all()


    def test_tilt_wavefront(self, create_wavefront: callable) -> None:
        """
        Tests the tilt_wavefront method.
        """
        wf = create_wavefront()

        # Test string inputs
        with pytest.raises(AssertionError):
            wf.tilt_wavefront("some string")

        # Test wrong shapes
        with pytest.raises(AssertionError):
            wf.tilt_wavefront(np.ones(1))

        # Test wrong shapes
        with pytest.raises(AssertionError):
            wf.tilt_wavefront(np.array(1.))

        # Test wrong shapes
        with pytest.raises(AssertionError):
            wf.tilt_wavefront(np.ones(3))

        # Test basic behaviour
        wf.tilt_wavefront(np.ones(2))


    def test_multiply_amplitude(self, create_wavefront: callable) -> None:
        """
        Tests the multiply_amplitude method.
        """
        wf = create_wavefront()

        # Test string inputs
        with pytest.raises(AssertionError):
            wf.multiply_amplitude("some string")

        # Test wrong shapes
        with pytest.raises(AssertionError):
            wf.multiply_amplitude(np.ones(1))

        # Test wrong shapes
        with pytest.raises(AssertionError):
            wf.multiply_amplitude(np.ones((1, 1, 1, 1)))

        # Test basic behaviour
        npix = wf.npixels
        wf.multiply_amplitude(np.array(1.))
        wf.multiply_amplitude(np.ones((npix, npix)))
        wf.multiply_amplitude(np.ones((1, npix, npix)))


    def test_add_phase(self, create_wavefront: callable) -> None:
        """
        Tests the add_phase method.
        """
        wf = create_wavefront()

        # Test string inputs
        with pytest.raises(AssertionError):
            wf.add_phase("some string")

        # Test wrong shapes
        with pytest.raises(AssertionError):
            wf.add_phase(np.ones(1))

        # Test wrong shapes
        with pytest.raises(AssertionError):
            wf.add_phase(np.ones((1, 1, 1, 1)))

        # Test basic behaviour
        npix = wf.npixels
        wf.add_phase(np.array(1.))
        wf.add_phase(np.ones((npix, npix)))
        wf.add_phase(np.ones((1, npix, npix)))


    def test_add_opd(self, create_wavefront: callable) -> None:
        """
        Tests the add_opd method.
        """
        wf = create_wavefront()

        # Test string inputs
        with pytest.raises(AssertionError):
            wf.add_opd("some string")

        # Test wrong shapes
        with pytest.raises(AssertionError):
            wf.add_opd(np.ones(1))

        # Test wrong shapes
        with pytest.raises(AssertionError):
            wf.add_opd(np.ones((1, 1, 1, 1)))

        # Test basic behaviour
        npix = wf.npixels
        wf.add_opd(np.array(1.))
        wf.add_opd(np.ones((npix, npix)))
        wf.add_opd(np.ones((1, npix, npix)))


    def test_normalise(self, create_wavefront: callable) -> None:
        """
        Tests the normalise method.
        """
        wf = create_wavefront()

        new_wf = wf.normalise()
        assert np.sum(new_wf.amplitude**2) == 1.


    def test_wavefront_to_psf(self, create_wavefront: callable) -> None:
        """
        Tests the wavefront_to_psf method.
        """
        wf = create_wavefront()
        wf.wavefront_to_psf()


    def test_invert_x_and_y(self, create_wavefront: callable) -> None:
        """
        Tests the invert_x_and_y method.
        """
        wf = create_wavefront()

        flipped_ampl = np.flip(wf.amplitude, axis=(-1, -2))
        flipped_phase = np.flip(wf.phase, axis=(-1, -2))
        new_wf = wf.invert_x_and_y()
        assert (new_wf.amplitude == flipped_ampl).all()
        assert (new_wf.phase == flipped_phase).all()


    def test_invert_x(self, create_wavefront: callable) -> None:
        """
        Tests the invert_x method.
        """
        wf = create_wavefront()

        flipped_ampl = np.flip(wf.amplitude, axis=-1)
        flipped_phase = np.flip(wf.phase, axis=-1)
        new_wf = wf.invert_x()
        assert (new_wf.amplitude == flipped_ampl).all()
        assert (new_wf.phase == flipped_phase).all()


    def test_invert_y(self, create_wavefront: callable) -> None:
        """
        Tests the invert_y method.
        """
        wf = create_wavefront()

        flipped_ampl = np.flip(wf.amplitude, axis=-2)
        flipped_phase = np.flip(wf.phase, axis=-2)
        new_wf = wf.invert_y()
        assert (new_wf.amplitude == flipped_ampl).all()
        assert (new_wf.phase == flipped_phase).all()


    def test_interpolate(self, create_wavefront: callable) -> None:
        """
        Tests the interpolate method.
        """
        wf = create_wavefront()
        wf = dLux.CircularAperture(1.)(wf)

        npix = wf.npixels
        pixscale = wf.pixel_scale
        k = 2
        new_wf = wf.interpolate(npix//k, pixscale*k)
        new_wf2 = wf.interpolate(npix//k, pixscale*k, real_imaginary=True)
        small_ampl = dLux.IntegerDownsample(k)(wf.amplitude[0])/k**2

        assert np.allclose(new_wf.amplitude[0], small_ampl)
        assert np.allclose(new_wf2.amplitude[0], small_ampl)


    def test_rotate(self, create_wavefront: callable) -> None:
        """
        Tests the rotate method.
        """
        wf = create_wavefront()
        wf = dLux.CircularAperture(1.)(wf)
        flipped_amplitude = np.flip(wf.amplitude, axis=(-1, -2))
        flipped_phase = np.flip(wf.phase, axis=(-1, -2))

        new_wf = wf.rotate(np.pi, order=1)
        assert np.allclose(new_wf.amplitude, flipped_amplitude, atol=1e-5)
        assert np.allclose(new_wf.phase, flipped_phase)

        new_wf = wf.rotate(np.pi, real_imaginary=True, order=1)

        assert np.allclose(new_wf.amplitude, flipped_amplitude, atol=1e-5)
        # Add small remainer to fix 0-pi instability
        assert np.allclose((new_wf.phase+1e-6)%np.pi, flipped_phase, atol=1e-5)

        with pytest.raises(NotImplementedError):
            wf.rotate(np.pi, fourier=True)


    def test_pad_to(self, create_wavefront: callable) -> None:
        """
        Tests the pad_to method.
        """
        even_wf = create_wavefront()
        odd_wf = create_wavefront(amplitude=np.ones((1, 15, 15)),
                                        phase=np.ones((1, 15, 15)))

        # Smaller value
        with pytest.raises(AssertionError):
            even_wf.pad_to(14)

        # even -> odd
        with pytest.raises(AssertionError):
            even_wf.pad_to(17)

        # odd -> even
        with pytest.raises(AssertionError):
            odd_wf.pad_to(16)

        assert even_wf.pad_to(20).npixels == 20


    def test_crop_to(self, create_wavefront: callable) -> None:
        """
        Tests the crop_to method.
        """
        even_wf = create_wavefront()
        odd_wf = create_wavefront(amplitude=np.ones((1, 15, 15)),
                                        phase=np.ones((1, 15, 15)))

        # Smaller value
        with pytest.raises(AssertionError):
            even_wf.crop_to(18)

        # even -> odd
        with pytest.raises(AssertionError):
            even_wf.crop_to(15)

        # odd -> even
        with pytest.raises(AssertionError):
            odd_wf.crop_to(14)

        assert even_wf.crop_to(12).npixels == 12



from jax import numpy as np 


__all__ = ["validate_eq_attr_dims", "validate_bc_attr_dims"]


bc_attr_dims_message = """
Expected dimensions that could be broadcast with
{}. Recieved dimensions {} when initialising attribute {}.
"""


eq_attr_dims_message = """
Expected dimension {} but recieved {} when initialising the attribute {}.
"""


class DimensionError(Exception):
    """
    This error is raised in the place of a ValueError
    when a user attempts to initialise a dLux object 
    with an attribute that has incorrect or inconsistent
    dimensions. 
    """


    def __init__(self: object, message: str) -> object:
        """
        Parameters
        ----------
        message: str
            A detailed, situational error message that is 
            correctly formatted. 
        """
        super().__init__(message)


def validate_eq_attr_dims(
        attr_shape: tuple, 
        correct_shape: tuple, 
        attr_name: str) -> None:
    """
    Confirm that an initialised attribute has the correct shape
    and raise a helpful error if it does not. Correct in this 
    case implies equality between `attr_shape` and `correct_shape`.

    Parameters
    ----------
    attr_shape: tuple
        The shape of the attribute to be initialised.
    correct_shape: tuple
        A shape that meets the necessary conditions of
        the attribute. 
    attr_name: str
        The name of the attribute that is getting 
        initialised
    """
    if attr_shape != correct_shape:
        raise DimensionError(
            eq_attr_dims_message.format(
                correct_shape, attr_shape, attr_name))


def validate_bc_attr_dims(
        attr_shape: tuple, 
        correct_shape: tuple,
        attr_name: str) -> None:
    """
    Confirm that an initialised attribute has the correct shape
    and raise a helpful error if it does not. Correct in this 
    case is the relaxed "broadcastable" version.

    Parameters
    ----------
    attr_shape: tuple
        The shape of the attribute to be initialised.
    correct_shape: tuple
        A shape that meets the necessary conditions of
        the attribute. 
    attr_name: str
        The name of the attribute that is getting 
        initialised
    """
    try:
        np.broadcast_shapes(attr_shape, correct_shape)
    except ValueError:
        raise DimensionError(
            bc_attr_dims_message.format(
                correct_shape, attr_shape, attr_name))






import pytest
import sys
import os

platform: str = sys.platform
not_lin_err: str = """
    The installation test is designed to run on
    a linux machine. In particular, this test
    is designed to be run by github actions.
    If you are using windows dispare.
"""

if platform not in ["linux", "linux2"]:
    print(f"Error: {not_lin_err}")


def test_install_dLux():
    install_ok: int = os.system("pip install --quiet .")

    assert install_ok == 0

    os.system("rm -r dLux.egg-info")
    os.system("rm -r build")


def test_import_dLux():
    import dLux
    assert True
    





from __future__ import annotations
import jax.numpy as np
import pytest
import dLux
from jax import config
config.update("jax_debug_nans", True)


Array = np.ndarray


class TestSource():
    """
    Tests the Source class.
    """


    def test_constructor(self, create_source : callable) -> None:
        """
        Test the constructor class.
        """
        # Position
        # Test string inputs
        with pytest.raises(ValueError):
            create_source(position="")

        # Test zero dimension input
        with pytest.raises(AssertionError):
            create_source(position=5.)

        # Test zero length input
        with pytest.raises(AssertionError):
            create_source(position=[])

        # Flux
        # Test string inputs
        with pytest.raises(ValueError):
            create_source(flux="")

        # Test 1d dimension input
        with pytest.raises(AssertionError):
            create_source(flux=[1.])

        # Test zero length input
        with pytest.raises(AssertionError):
            create_source(flux=[])

        # Spectrum
        # Test non-spectrum input
        with pytest.raises(AssertionError):
            create_source(spectrum=[])

        # Name
        # Test non-string input
        with pytest.raises(AssertionError):
            create_source(name=[])


    # Getters
    def test_get_position(self, create_source : callable) -> None:
        """
        Tests the get_position method.
        """
        source = create_source()
        assert (source.get_position() == source.position).all()


    def test_get_flux(self, create_source : callable) -> None:
        """
        Tests the get_flux method.
        """
        source = create_source()
        assert (source.get_flux() == source.flux).all()


    def test_get_wavelengths(self, create_source : callable) -> None:
        """
        Tests the get_wavelengths method.
        """
        source = create_source()
        assert (source.get_wavelengths() == source.spectrum.wavelengths).all()


    def test_get_weights(self, create_source : callable) -> None:
        """
        Tests the get_weights method.
        """
        source = create_source()
        assert (source.get_weights() == source.spectrum.weights).all()

class TestResolvedSource():
    """
    Tests the ResolvedSourve class.
    """
    pass


class TestRelativeFluxSource():
    """
    Tests the RelativeFluxSource class.
    """


    def test_constructor(self, create_relative_flux_source : callable) -> None:
        """
        Tests the constructor.
        """
        # Test string inputs
        with pytest.raises(ValueError):
            create_relative_flux_source(contrast="")

        # Test one dimension input
        with pytest.raises(AssertionError):
            create_relative_flux_source(contrast=[5.])

        # Test zero length input
        with pytest.raises(AssertionError):
            create_relative_flux_source(contrast=[])


    def test_get_flux(self, create_relative_flux_source : callable) -> None:
        """
        Tests the get_flux method.
        """
        source = create_relative_flux_source()
        flux_out = source.get_flux()
        assert flux_out.shape == (2,)
        assert np.allclose(flux_out[0]/flux_out[1], source.contrast)


class TestRelativePositionSource():
    """
    Tests the RelativePositionSource class.
    """


    def test_constructor(self, create_relative_position_source : callable) -> None:
        """
        Tests the constructor.
        """
        # Separation
        # Test string inputs
        with pytest.raises(ValueError):
            create_relative_position_source(separation="")

        # Test one dimension input
        with pytest.raises(AssertionError):
            create_relative_position_source(separation=[5.])

        # Test zero length input
        with pytest.raises(AssertionError):
            create_relative_position_source(separation=[])

        # position_angle
        # Test string inputs
        with pytest.raises(ValueError):
            create_relative_position_source(position_angle="")

        # Test one dimension input
        with pytest.raises(AssertionError):
            create_relative_position_source(position_angle=[5.])

        # Test zero length input
        with pytest.raises(AssertionError):
            create_relative_position_source(position_angle=[])


    def test_get_position(self, create_relative_position_source : callable) -> None:
        """
        Tests the get_position method.
        """
        source = create_relative_position_source()
        position_out = source.get_position()
        sep_vec = position_out[0] - position_out[1]
        separation = np.hypot(sep_vec[0], sep_vec[1])
        position_angle = np.arctan2(sep_vec[0], sep_vec[1])
        assert position_out.shape == (2,2)
        assert np.allclose(source.separation, separation).all()
        assert np.allclose(source.position_angle, position_angle).all()


class TestPointSource():
    """
    Tests the PointSource class.
    """


    def test_model(self, create_point_source : callable) -> None:
        """
        Tests the model method.
        """
        # TODO: this has no asserts?
        source = create_point_source()
        optics = dLux.core.Optics([dLux.CreateWavefront(16, 1)])
        detector = dLux.core.Detector([dLux.AddConstant(0.)])
        # filter_in = dLux.Filter()
        source.model(optics)
        source.model(optics, detector)
        # source.model(optics, detector, filter_in)
        # source.model(optics, filter_in=filter_in)


class TestMultiPointSource():
    """
    Tests the MultiPointSource class.
    """


    def test_constructor(self, create_multi_point_source : callable) -> None:
        """
        Test the constructor class.
        """
        # Position
        # Test string inputs
        with pytest.raises(ValueError):
            create_multi_point_source(position="")

        # Test zero dimension input
        with pytest.raises(AssertionError):
            create_multi_point_source(position=5.)

        # Test zero length input
        with pytest.raises(AssertionError):
            create_multi_point_source(position=[])

        # Test 1 dim input
        with pytest.raises(AssertionError):
            create_multi_point_source(position=np.ones(2))

        # Flux
        # Test string inputs
        with pytest.raises(ValueError):
            create_multi_point_source(flux="")

        # Test zero length input
        with pytest.raises(AssertionError):
            create_multi_point_source(flux=[])

        # Test 2 dim input
        with pytest.raises(AssertionError):
            create_multi_point_source(flux=np.ones((2, 2)))

        # Spectrum
        # Test non-spectrum input
        with pytest.raises(AssertionError):
            create_multi_point_source(spectrum=[])

        # Name
        # Test non-string input
        with pytest.raises(AssertionError):
            create_multi_point_source(name=[])


    def test_model(self, create_multi_point_source : callable) -> None:
        """
        Tests the model method.
        """
        source = create_multi_point_source()
        optics = dLux.core.Optics([dLux.CreateWavefront(16, 1)])
        detector = dLux.core.Detector([dLux.AddConstant(0.)])
        # filter_in = dLux.Filter()
        source.model(optics)
        source.model(optics, detector)
        # source.model(optics, detector, filter_in)
        # source.model(optics, filter_in=filter_in)


class TestArrayDistribution():
    """
    Tests the ArrayDistribution class.
    """


    def test_constructor(self, create_array_distribution : callable) -> None:
        """
        Tests the constructor.
        """
        # Test string inputs
        with pytest.raises(ValueError):
            create_array_distribution(distribution="")

        # Test one dimension input
        with pytest.raises(AssertionError):
            create_array_distribution(distribution=[5.])

        # Test zero length input
        with pytest.raises(AssertionError):
            create_array_distribution(distribution=[])


    def test_get_distribution(self, create_array_distribution : callable) -> None:
        """
        Tests the get_distribution method.
        """
        source = create_array_distribution()
        assert (source.get_distribution() == source.distribution).all()


    def test_normalise(self, create_array_distribution : callable) -> None:
        """
        Tests the normalise method.
        """
        source = create_array_distribution()
        new_distribution = np.ones((4, 4))
        new_source = source.set('distribution', new_distribution).normalise()
        assert np.allclose(new_source.distribution.sum(), 1.)


    def test_model(self, create_array_distribution : callable) -> None:
        """
        Tests the model method.
        """
        source = create_array_distribution()
        optics = dLux.core.Optics([dLux.CreateWavefront(16, 1)])
        detector = dLux.core.Detector([dLux.AddConstant(0.)])
        # filter_in = dLux.Filter()
        source.model(optics)
        source.model(optics, detector)
        # source.model(optics, detector, filter_in)
        # source.model(optics, filter_in=filter_in)


class TestBinarySource():
    """
    Tests the BinarySource class.
    """


    def test_model(self, create_binary_source : callable) -> None:
        """
        Tests the model method.
        """
        # TODO add tests
        source = create_binary_source()
        optics = dLux.core.Optics([dLux.CreateWavefront(16, 1)])
        detector = dLux.core.Detector([dLux.AddConstant(0.)])
        # filter_in = dLux.Filter()
        source.model(optics)
        source.model(optics, detector)
        # source.model(optics, detector, filter_in)
        # source.model(optics, filter_in=filter_in)

class TestPointExtendedSource():
    """
    Tests the PointExtendedSource class.
    """

    def test_model(self, create_point_extended_source : callable) -> None:
        """
        Tests the model method.
        """
        source = create_point_extended_source()
        optics = dLux.core.Optics([dLux.CreateWavefront(16, 1)])
        detector = dLux.core.Detector([dLux.AddConstant(0.)])
        # filter_in = dLux.Filter()
        source.model(optics)
        source.model(optics, detector)
        # source.model(optics, detector, filter_in)
        # source.model(optics, filter_in=filter_in)



class TestPointAndExtendedSource():
    """
    Tests the PointAndExtendedSource class.
    """


    def test_constructor(self, create_point_and_extended_source : callable) -> None:
        """
        Tests the constructor.
        """
        # Test non Combined Spectrum input
        with pytest.raises(AssertionError):
            spec = dLux.spectrums.ArraySpectrum(np.linspace(500e-9, 600e-9, 10))
            create_point_and_extended_source(spectrum=spec)


    def test_model(self, create_point_and_extended_source : callable) -> None:
        """
        Tests the model method.
        """
        source = create_point_and_extended_source()
        optics = dLux.core.Optics([dLux.CreateWavefront(16, 1)])
        detector = dLux.core.Detector([dLux.AddConstant(0.)])
        # filter_in = dLux.Filter()
        source.model(optics)
        source.model(optics, detector)
        # source.model(optics, detector, filter_in)
        # source.model(optics, filter_in=filter_in)



from __future__ import annotations
from zodiax import Base
from jax.tree_util import tree_map
import jax.numpy as np
from jax import vmap
from abc import abstractmethod
from equinox import tree_at
from typing import Any
import dLux


Array = np.ndarray


__all__ = ["AbstractObservation", "Dither"]


class AbstractObservation(Base):
    """
    Abstract base class for observations. All observations should inherit from
    this class and must implement an `.observe()` method that only takes in a
    single instance of `dLux.Instrument`.

    Attributes
    ----------
    name : str
        The name of the observation that can be used to access the observation
        from the `Instrument` class.
    """
    name : str


    def __init__(self : AbstractObservation, 
                 name : str = 'AbstractObservation'):
        """
        Constructor for the AbstractObservation class.

        Parameters
        ----------
        name : str = 'AbstractObservation'
            The name of the observation that can be used to access the
            observation from the `Instrument` class.
        """
        self.name = str(name)


    @abstractmethod
    def observe(self       : AbstractObservation, 
                instrument : dLux.core.Instrument) -> Any:
        """
        Abstract method for the observation function.
        """
        pass


class Dither(AbstractObservation):
    """
    Observation class designed to apply a series of dithers to the insturment
    and return the corresponding psfs.

    Attributes
    ----------
    dithers : Array, (radians)
        The array of dithers to apply to the source positions. The shape of the
        array should be (ndithers, 2) where ndithers is the number of dithers
        and the second dimension is the (x, y) dither in radians.
    name : str
        The name of the observation that can be used to access the observation
        from the `Instrument` class.
    """
    dithers : Array


    def __init__(self : Dither, dithers : Array, name : str = 'Dither'):
        """
        Constructor for the Dither class.

        Parameters
        ----------
        dithers : Array, (radians)
            The array of dithers to apply to the source positions. The shape of
            the array should be (ndithers, 2) where ndithers is the number of
            dithers and the second dimension is the (x, y) dither in radians.
        name : str = 'Dither'
            The name of the observation that can be used to access the
            observation from the `Instrument` class.
        """
        super().__init__(name)
        self.dithers = np.asarray(dithers, float)
        dLux.exceptions.validate_bc_attr_dims(self.dithers.shape, (1, 2), 
            'dithers')


    def dither_position(self       : Dither, 
                        instrument : Instrument, 
                        dither     : Array) -> Instrument:
        """
        Dithers the position of the source objects by dither.

        Parameters
        ----------
        dither : Array, radians
            The (x, y) dither to apply to the source positions.

        Returns
        -------
        instrument : Instrument
            The instrument with the sources dithered.
        """
        assert dither.shape == (2,), ("dither must have shape (2,) ie (x, y)")

        # Define the dither function
        dither_fn = lambda source: source.add('position', dither)

        # Map the dithers across the sources
        dithered_sources = tree_map(dither_fn, instrument.sources, \
            is_leaf = lambda leaf: isinstance(leaf, dLux.sources.Source))

        # Apply updates
        return tree_at(lambda instrument: instrument.sources, instrument, 
            dithered_sources)


    def observe(self       : Dither,
                instrument : Instrument,
                *args, 
                **kwargs) -> Array:
        """
        Applies a series of dithers to the instrument sources and calls the
        .model() method after applying each dither.

        Parameters
        ----------
        instrument : Instrument
            The array of dithers to apply to the source positions.

        Returns
        -------
        psfs : Array
            The psfs generated after applying the dithers to the source
            positions.
        """
        dith_fn = lambda dither: self.dither_position(instrument, 
            dither).model(*args, **kwargs)
        return vmap(dith_fn, 0)(self.dithers)